import pytest
import jax
import jax.numpy as jnp
from pymob import Config

# this should raise an exception when nans are created the computation

def init_guts_casestudy_constant_exposure(scenario="testing_guts_constant_exposure"):
    """This is an external case study used for local testing. The test study
    will eventually added also to the remote as an example, but until this happens
    the test is skipped on the remote.
    """
    config = Config()
    config.case_study.name = "guts_base"
    config.case_study.scenario = scenario
    config.case_study.package = "../pollinator-tktd/case_studies"
    config.case_study.simulation = "GutsSimulationConstantExposure"
    config.import_casestudy_modules(reset_path=True)

    if "sim" in config._modules:       
        Simulation = config.import_simulation_from_case_study()
        sim = Simulation(config)
        return sim
    else:
        pytest.skip()


def init_guts_casestudy_variable_exposure(scenario="testing_guts_variable_exposure"):
    """This is an external case study used for local testing. The test study
    will eventually added also to the remote as an example, but until this happens
    the test is skipped on the remote.
    """
    config = Config()
    config.case_study.name = "guts_base"
    config.case_study.scenario = scenario
    config.case_study.package = "../pollinator-tktd/case_studies"
    config.case_study.simulation = "GutsSimulationVariableExposure"
    config.import_casestudy_modules(reset_path=True)

    if "sim" in config._modules:       
        Simulation = config.import_simulation_from_case_study()
        sim = Simulation(config)
        return sim
    else:
        pytest.skip()

@pytest.mark.slow
def test_guts_constant_exposure():
    sim = init_guts_casestudy_constant_exposure()
    sim.initialize_from_script()

    sim.use_jax_solver()
    sim.dispatch_constructor(rtol=1e-7, atol=1e-8)

    # test gradients
    def test_func(theta):
        k_d, h_b, b, z = theta
        e = sim.dispatch({"k_d": k_d, "h_b": h_b, "b":b, "z": z})
        e()
        return e.Y["survival"][-1]

    val, grads = jax.value_and_grad(test_func)(jnp.array([1, 0.001, 0.2, 1.0]))

    assert all(grads != jnp.nan)

@pytest.mark.slow
def test_guts_variable_exposure():
    sim = init_guts_casestudy_variable_exposure()
    sim.load_modules()
    sim.initialize_from_script()

    sim.use_jax_solver()
    sim.dispatch_constructor(rtol=1e-7, atol=1e-8)

    # test gradients
    def test_func(theta):
        k_d, h_b, b, z = theta
        e = sim.dispatch({"k_d": k_d, "h_b": h_b, "b":b, "z": z})
        e()
        return e.Y["survival"][-1]

    val, grads = jax.value_and_grad(test_func)(jnp.array([1, 0.001, 0.2, 1.0]))

    assert all(grads != jnp.nan)



if __name__ == "__main__":
    test_guts_variable_exposure()


from functools import partial
from pathlib import Path

import jax
import jax.numpy as jnp
import sympy as sp
import numpy as np
from scipy.interpolate import interp1d

from pymob.solvers.base import mappar
from pymob.solvers.symbolic import (
    PiecewiseSymbolicODESolver, FunctionPythonCode, get_return_arguments, dX_dt2X
)

class RED_SD:
    """Simplest guts model, mainly for testing"""
    @staticmethod
    def _rhs_jax(t, y, x_in, kd, b, m, hb):
        D, H = y
        dD_dt = kd * (x_in.evaluate(t) - D)
        dH_dt = b * jnp.maximum(0.0, D - m) + hb
        return dD_dt, dH_dt

    @staticmethod
    def _solver_post_processing(results, t, interpolation):
        results["survival"] = jnp.exp(-results["H"])
        results["exposure"] = jax.vmap(interpolation.evaluate)(t)
        return results

red_sd = RED_SD._rhs_jax
red_sd_post_processing = RED_SD._solver_post_processing

class RED_SD_DA(RED_SD): 
    @staticmethod
    def _rhs_jax(t, y, x_in, kd, w, b, m, hb):
        D, H = y
        dD_dt = kd * (x_in.evaluate(t) - D)
        dH_dt = b * jnp.maximum(0.0, jnp.sum(w * D) - m) + hb
        return dD_dt, dH_dt

red_sd_da = RED_SD_DA._rhs_jax
red_sd_da_post_processing = RED_SD_DA._solver_post_processing


def guts_constant_exposure(t, y, C_0, k_d, z, b, h_b):
    # for constant exposure
    D, H, S = y
    dD_dt = k_d * (C_0 - D)
    
    switchDS = 0.5 + (1 / jnp.pi) * jnp.arctan(1e16 * (D - z))
    dH_dt = (b * switchDS * (D - z) + h_b)

    dS_dt = -dH_dt * S

    return dD_dt, dH_dt, dS_dt

def guts_variable_exposure(t, y, x_in, k_d, z, b, h_b):
    # for constant exposure
    D, H, S = y
    C = x_in.evaluate(t)
    dD_dt = k_d * (C - D)
    
    switchDS = 0.5 + (1 / jnp.pi) * jnp.arctan(1e16 * (D - z))
    dH_dt = (b * switchDS * (D - z) + h_b)

    dS_dt = -dH_dt * S

    return dD_dt, dH_dt, dS_dt


class Interpolation:
    def __init__(self, xs, ys, method="previous") -> None:
        self.f = interp1d(
            x=xs,
            y=ys,
            axis=0,
            kind=method
        )

    def evaluate(self, t) -> np.ndarray:
        return self.f(t)


class PiecewiseSymbolicSolver(PiecewiseSymbolicODESolver):
    interpolation_type = "previous"

    def t_jump(self, func_name, compiled_functions={}):
        t, Y, Y_0, theta = self.define_symbols()
        
        D_t = compiled_functions["F"]["algebraic_solutions"]["D"]
        z = theta["z"]
        eq = sp.Eq(D_t.rhs, z).expand()

        t_0 = sp.solve(eq, t)

        assert len(t_0) == 1
        func = sp.simplify(t_0[0])

        python_code = FunctionPythonCode(
            func_name=func_name,
            lhs_0=("Y_0", tuple(Y_0.keys())),
            theta=("theta", tuple(theta.keys())),
            lhs=("Y",),
            rhs=(func,),
            expand_arguments=False,
            modules=("numpy","scipy"),
            docstring=""
        )
        
        tex = self.to_latex(solutions=[sp.Eq(sp.Symbol(func_name), func)])
        code_file = Path(self.output_path, f"{func_name}.tex")
        with open(code_file, "w") as f:
            f.writelines(tex)

        return func, python_code

    def define_symbols(self):
        """Define the necessary symbols solely based on the function"""
        thetanames = mappar(
            self.model, {}, 
            exclude=["t", "dt", "y", "x_in", "Y", "X"], 
            to="names"
        )
        ynames = [dX_dt2X(a) for a in get_return_arguments(self.model)]

        # define symbols for t, Y, Y_0 and theta
        t = sp.Symbol("t", positive=True, real=True)
        Y = {y: sp.Function(y, positive=True, real=True) for y in ynames}
        Y_0 = {
            f"{y}_0": sp.Symbol(f"{y}_0", positive=True, real=True) 
            for y in ynames
        }
        theta = {p: sp.Symbol(p, positive=True, real=True) for p in thetanames}
        
        symbols = (t, Y, Y_0, theta)

        return symbols
        
    @property
    def compiler_recipe(self):
        return {
            "F":self.compile_model,
            "t_jump":self.t_jump,
            "F_piecewise":partial(self.jump_solution, funcnames="F t_jump")
        }

    def solve(self, parameters, y0, x_in):
        odeargs = mappar(
            self.model, 
            parameters, 
            exclude=["t", "dt", "y", "x_in"], 
            to="dict"
        )

        F_piecewise = self.compiled_functions["F_piecewise"]["compiled_function"]

        # get arguments
        time = np.array(self.x)
        Y_0_values = [v for v in y0.values()]

        # handle interpolation
        if "exposure" in self.coordinates_input_vars["x_in"]:
            exposure_interpolation = Interpolation(
                xs=self.coordinates_input_vars["x_in"]["exposure"][self.x_dim],
                ys=x_in["exposure"],
                method=self.interpolation_type
            )
        else:
            exposure_interpolation = Interpolation(
                xs=self.x,
                ys=np.full_like(self.x, odeargs["C_0"]), #type: ignore
                method=self.interpolation_type
            )

        # run main loop
        sol = [Y_0_values]
        for i in range(1, len(time)):
            # parse arguments
            C_0 = exposure_interpolation.evaluate(time[i-1])
            odeargs.update({"C_0": C_0}) # type: ignore
            dt = time[i] - time[i-1]

            # call piecewise function
            y_t = F_piecewise(
                t=dt, 
                Y_0=sol[i-1], 
                θ=tuple(odeargs.values()), # type: ignore
                ε=1e-14
            ) 
            
            sol.append(y_t)

        Y_t = np.array(sol)

        results = {k:y for k, y in zip(y0.keys(), Y_t.T)}
        results["exposure"] = exposure_interpolation.evaluate(np.array(self.x))

        return results


from . import sim
from . import mod
from . import data
from . import prob
from . import plot

__version__ = "1.0.0"

from .sim import (
    GutsBase,
    PymobSimulator,
    ECxEstimator,
    LPxEstimator, 
    GutsBaseError,
)


from functools import partial
import numpy as np
import jax
import jax.numpy as jnp
from numpyro.infer import Predictive
from typing import Literal


def survival_predictions(
    probs, n_trials, 
    eps: float = 0.0,
    seed=1, 
    mode: Literal["survival", "lethality", "deaths"] = "survival"
):
    """Generate predictions for survival based on a multinomial survival distribution

    Parameters
    ----------
    probs : ArrayLike
        2D Array denoting the multinomial probabilities of deaths for each time interval 
        per experiment 
        dims=(experiment, time)
    n_trials : ArrayLike
        1D Array denoting the number of organisms at the beginning in each experiment
        dims = (experiment,)
    seed : int, optional
        Seed for the random number generator, by default 1
    mode : str, optional
        How should the random draws be returned? 
        - survival: Decreasing from n_trials to 0
        - lethality: Increasing from 0 to n_trials
        - deaths: Between 0 and n_trials in each interval. Summing to n_trials
    """

    def survival_to_death_probs(pr_survival):
        # truncate here, because numeric errors below the solver tolerance can
        # lead to negative values in the difference. This needs to be cured here
        pr_survival_ = np.trunc(pr_survival / eps) * eps
        pr_death = pr_survival_[:-1] - pr_survival_[1:]

        pr_death = np.concatenate([
            # concatenate a zero at the beginning in order to "simulate" no
            # deaths at T = 0
            jnp.zeros((1,)),
            # Delta S
            pr_death,
            # The remaining mortility as T -> infinity 
            jnp.ones((1,))-pr_death.sum()
        ])

        # make sure the vector is not zero or 1 (this is always problematic for 
        # probabilities) and make sure the vector sums to 1
        pr_death = np.clip(pr_death, eps, 1-eps)
        pr_death = pr_death / pr_death.sum()
        return pr_death

    rng = np.random.default_rng(seed)
    deaths = jnp.array(list(map(
        lambda n, pr_survival: rng.multinomial(
            n=n, pvals=survival_to_death_probs(pr_survival)
        ), 
        n_trials, 
        probs
    )))

    # remove the last observations to trim off the simulated unobserved mortality
    deaths = deaths[:, :-1]

    if mode == "deaths":
        return deaths
    elif mode == "lethality":
        return deaths.cumsum(axis=1)
    elif mode == "survival":
        return np.expand_dims(n_trials, axis=1) - deaths.cumsum(axis=1)
    else:
        raise NotImplementedError(
            f"Mode {mode} is not implemented."+
            "Use one of 'survival', 'lethality', or 'deaths'."
        )
    
def posterior_predictions(sim, idata, seed=None):
    """Make posterior predictions for survival data"""
    if seed is None:
        seed = sim.config.simulation.seed

    n = idata.posterior.dims["draw"]
    c = idata.posterior.dims["chain"]

    key = jax.random.PRNGKey(seed)

    obs, masks = sim.inferer.observation_parser()

    model_kwargs = sim.inferer.preprocessing(obs=obs, masks=masks)
    if sim.config.inference_numpyro.user_defined_error_model is None:
        model_kwargs["obs"]["survival"] = None

    # prepare model
    model = partial(
        sim.inferer.inference_model, 
        solver=sim.inferer.evaluator, 
        **model_kwargs
    )    

    posterior_samples = {
        k: np.array(v["data"]) for k, v 
        in idata.unconstrained_posterior.to_dict()["data_vars"].items()
    }

    predictive = Predictive(
        model, posterior_samples=posterior_samples,
        num_samples=n, batch_ndims=2
    )

    samples = predictive(key)

    if sim.config.inference_numpyro.user_defined_error_model is not None:
        chains = []
        for i in range(c):
            # TODO: Switch to vmap and jit, but it did not work, so if you do it TEST IT!!!!!!!
            predictions = list(map(
                partial(
                    survival_predictions,
                    n_trials=obs["survival"][:, 0].astype(int),
                    eps=obs["eps"],
                    seed=seed,
                    mode="survival",
                ),
                samples["survival"][i]
            ))
            chains.append(predictions)

        posterior_predictive = {"survival_obs": np.array(chains)}
    else:
        posterior_predictive = {"survival_obs": samples.pop("survival_obs")}


    new_idata = sim.inferer.to_arviz_idata(
        posterior=samples,
        posterior_predictive=posterior_predictive,
        n_draws=n,
        n_chains=c
    )

    # update chain names in case they were subselected (clustering)
    # new_idata = new_idata.assign_coords({"chain": idata.posterior.chain.values})
    new_idata = new_idata.assign_coords({"chain": idata.posterior.chain.values})

    # assert the new posterior matches the old posterior
    tol = sim.config.jaxsolver.atol * 100
    abs_diff_posterior = np.abs(idata.posterior - new_idata.posterior)
    np.testing.assert_array_less(abs_diff_posterior.mean().to_array(), tol)

    fit_tol = tol * sim.coordinates["time"].max()
    abs_diff_fits = np.abs(new_idata.posterior_model_fits - idata.posterior_model_fits)
    np.testing.assert_array_less(abs_diff_fits.mean().to_array(), fit_tol)

    idata.posterior_model_fits = new_idata.posterior_model_fits
    idata.posterior_predictive = new_idata.posterior_predictive

    return idata


import pytest
from guts_base import GutsBase
from guts_base.sim import construct_sim_from_config

# run tests with the Simulation fixtures
def test_setup(sim):
    """Tests the construction method"""
    assert True


def test_simulation(sim):
    """Tests if a forward simulation pass can be computed"""
    sim.dispatch_constructor()
    evaluator = sim.dispatch()
    evaluator()
    evaluator.results

    assert True
            
def test_copy(sim):
    sim.dispatch_constructor()
    e_orig = sim.dispatch()
    e_orig()
    e_orig.results

    sim_copy = sim.copy()
    
    sim_copy.dispatch_constructor()
    e_copy = sim_copy.dispatch()
    e_copy()

    assert (e_copy.results == e_orig.results).all().to_array().all().values


@pytest.mark.slow
@pytest.mark.parametrize("backend", ["numpyro"])
def test_inference(sim: GutsBase, backend):
    """Tests if prior predictions can be computed for arbitrary backends"""
    sim.dispatch_constructor()
    sim.set_inferer(backend)

    sim.config.inference.n_predictions = 2
    sim.prior_predictive_checks()
    
    sim.config.inference_numpyro.kernel = "svi"
    sim.config.inference_numpyro.svi_iterations = 10
    sim.config.inference_numpyro.svi_learning_rate = 0.05
    sim.config.inference_numpyro.draws = 10
    sim.config.inference.n_predictions = 10

    sim.inferer.run()

    sim.posterior_predictive_checks()

    sim.config.report.debug_report = True
    sim.report()


if __name__ == "__main__":
    # test_inference(sim=construct_sim_from_config("red_sd", GutsBase), backend="numpyro")
    test_inference(sim=construct_sim_from_config("red_sd_da", GutsBase), backend="numpyro")



import warnings

def datalad_locked_file_warning(file):
    warnings.warn(
        f"The file '{file}' does not have write access. "
        f"To unlock the file, use DataLad with the command: "
        f"datalad unlock '{file}'."
    )


import warnings
from functools import partial
import numpy as np
import xarray as xr
from typing import Literal, Optional, Dict, List
import pandas as pd
from scipy.optimize import minimize
from matplotlib import pyplot as plt
from tqdm import tqdm

from pymob import SimulationBase
from guts_base.sim.utils import GutsBaseError

class ECxEstimator:
    """Estimates the exposure level that corresponds to a given effect. The algorithm 
    operates by varying a given exposure profile (x_in). For each new estimation, a new
    estimator is initialized.

    Parameters
    ----------

    sim : SimulationBase
        This must be a pymob.SimulationBase object. If the ECxEstimator.estimate method
        is used with the modes 'draw' or 'mean'    
    
    """
    _name = "EC"
    _parameter_msg = (
        "Manual estimation (mode='manual', without using posterior information) requires " +
        "specification of parameters={...}. You can obtain and modify " +
        "parameters using the pymob API: `sim.config.model_parameters.value_dict` " +
        "returns a dictionary of DEFAULT PARAMETERS that you can customize to your liking " +
        "(https://pymob.readthedocs.io/en/stable/api/pymob.sim.html#pymob.sim.config.Modelparameters.value_dict)."
    )                

    def __init__(
        self, 
        sim: SimulationBase, 
        effect: str, 
        x: float,
        time: float, 
        x_in: xr.Dataset, 
    ):
        self.sim = sim.copy()
        self.time = time
        self.x = x
        self.effect = effect
        self._mode = None

        # creates an empty observation dataset with the coordinates of the
        # original observations (especially time), except the ID, which is overwritten
        # and taken from the x_in dataset
        pseudo_obs = self.sim.observations.isel(id=[0])
        pseudo_obs = pseudo_obs.drop(["exposure","survival"])
        pseudo_obs["id"] = x_in["id"]
        
        self.sim.config.data_structure.survival.observed = False
        self.sim.observations = pseudo_obs

        # ensure correct coordinate order with x_in and raise errors early
        self.sim.model_parameters["x_in"] = self.sim.parse_input("x_in", x_in)


        # fix time after observations have been set. The outcome of the simulation
        # can dependend on the time vector, because in e.g. IT models, the time resolution
        # is important. Therefore the time at which the ECx is computed is just inserted
        # into the time vector at the right position.
        self.sim.coordinates["time"] = np.unique(np.concatenate([
            self.sim.coordinates["time"], np.array(time, ndmin=1)
        ]))

        self.sim.model_parameters["y0"] = self.sim.parse_input("y0", drop_dims="time")
        self.sim.dispatch_constructor()

        self.results = pd.Series({
            "mean": np.nan,
            "q05": np.nan,
            "q95": np.nan,
            "std": np.nan,
            "cv": np.nan,
            "msg": np.nan
        })

        self.figure_profile_and_effect = None
        self.figure_loss_curve = None


    def _assert_posterior(self):
        try:
            p = self.sim.inferer.idata.posterior
        except AttributeError:
            raise GutsBaseError(
                "Using mode='mode' or mode='draws', but sim did not contain a posterior. " + 
                "('sim.inferer.idata.posterior'). " + self._parameter_msg
            )


    
    def _evaluate(self, factor, theta):
        evaluator = self.sim.dispatch(
            theta=theta, 
            x_in=self.sim.validate_model_input(self.sim.model_parameters["x_in"] * factor)
        )
        evaluator()
        return evaluator

    def _loss(self, log_factor, theta):
        # exponentiate the log factor
        factor = np.exp(log_factor)

        e = self._evaluate(factor, theta)
        s = e.results.sel(time=self.time)[self.effect].values

        return (s - (1 - self.x)) ** 2

    def _posterior_mean(self):
        mean = self.sim.inferer.idata.posterior.mean(("chain", "draw"))
        mean = {k: v["data"] for k, v in mean.to_dict()["data_vars"].items()}
        return mean

    def _posterior_sample(self, i):
        posterior_stacked = self.sim.inferer.idata.posterior.stack(
            sample=("chain", "draw")
        )
        sample = posterior_stacked.isel(sample=i)
        sample = {k: v["data"] for k, v in sample.to_dict()["data_vars"].items()}
        return sample

    def plot_loss_curve(self, 
        mode: Literal["draws", "mean", "manual"] = "draws", 
        draws: Optional[int] = None, 
        parameters: Optional[Dict[str,float|List[float]]] = None,
        log_x0: float = 0.0, 
        force_draws: bool = False
    ):
        """
        Parameters
        ----------

        mode : Literal['draws', 'mean', 'manual']
            mode of estimation. mode='mean' takes the mean of the posterior and estimate
            the ECx for this singular value. mode='draws' takes samples from the posterior
            and estimate the ECx for each of the parameter draws. mode='manual' takes
            a parameter set (Dict) in the parameters argument and uses that for estimation. 
            Default: 'draws'
        
        draws : int
            Number of draws to take from the posterior. Only takes effect if mode='draw'.
            Raises an exception if draws < 100, because this is insufficient for a 
            reasonable uncertainty estimate. Default: None (using all samples from the
            posterior)
        
        parameters : Dict[str,float|list[float]]
            a parameter dictionary passed used as model parameters for finding the ECx
            value. Default: None

        
        log_x0 : float
            the starting value for the multiplication factor of the exposure profile for
            the minimization algorithm. This value is on the log scale. This means, 
            exp(log_x0=0.0) = 1.0, which means that the log_x0=0.0 will start at an
            unmodified exposure profile. Default: 0.0
        
        force_draws : bool
            Force the estimate method to accept a number of draws less than 100. Default: False

        """
        draws = self._check_mode_and_draws_and_parameters(mode, draws, parameters, force_draws)


        factor = np.linspace(-2,2, 100) + log_x0
        fig, ax = plt.subplots(1,1, sharey=True, figsize=(4, 3))
        
        for i in tqdm(range(draws)):
            if mode == "draws":
                sample = self._posterior_sample(i)
            elif mode == "mean":
                sample = self._posterior_mean()
            elif mode == "manual":
                sample = parameters
            else: 
                raise NotImplementedError(
                    f"Bad mode: {mode}. Mode must be one 'mean' or 'draws'"
                )
            
            y = list(map(partial(self._loss, theta=sample), factor))

            ax.plot(
                np.exp(factor), y, 
                color="black",
            )

        ax.plot(
            [], [], color="black",
            label=f"$\ell = S(t={self.time},x_{{in}}=C_{{ext}} \cdot \phi) - {self.x}$"
        )
        ax.set_ylabel("Loss ($\ell$)")
        ax.set_xlabel("Multiplication factor ($\phi$)")
        ax.set_title(f"ID: {self.sim.coordinates['id'][0]}")
        ax.set_ylim(0, ax.get_ylim()[1] * 1.25)
        ax.legend(frameon=False)
        fig.tight_layout()

        self.figure_loss_curve = fig

    def _check_mode_and_draws_and_parameters(self, mode, draws, parameters, force_draws):

        if mode == "draws":
            self._assert_posterior()

            if draws is None:
                draws = (
                    self.sim.inferer.idata.posterior.sizes["chain"] * 
                    self.sim.inferer.idata.posterior.sizes["draw"]
                )
            elif draws < 100 and not force_draws:
                raise GutsBaseError(
                    "draws must be larger than 100. Preferably > 1000. " +
                    f"If you don't want uncertainty assessment of the {self._name} " +
                    "estimates, use mode='mean'. If you really want to use less than " +
                    "100 draws, use force_draws=True at your own risk."
                )
            else:
                pass
            
            warnings.warn(
                "Values passed to 'parameters' don't have an effect in mode='draws'"
            )

        elif mode == "mean":
            self._assert_posterior()

            draws = 1

            warnings.warn(
                "Values passed to 'parameters' don't have an effect in mode='draws'"
            )

        elif mode == "manual":
            draws = 1
            if parameters is None:
                raise GutsBaseError(self._parameter_msg)
        else:
            raise GutsBaseError(
                f"Bad mode: {mode}. Mode must be one 'mean' or 'draws'"
            )

        return draws


    def estimate(
        self, 
        mode: Literal["draws", "mean", "manual"] = "draws", 
        draws: Optional[int] = None, 
        parameters: Optional[Dict[str,float|List[float]]] = None,
        log_x0: float = 0.0, 
        accept_tol: float = 1e-5, 
        optimizer_tol: float = 1e-5,
        method: str = "cobyla", 
        show_plot: bool = True,
        force_draws: bool = False,
        **optimizer_kwargs
    ):
        """The minimizer for the EC_x operates on the unbounded linear scale, estimating 
        the log-modification factor. Converted to the linear scale by factor=exp(x), the 
        profile modification factor is obtained.

        Using x0=0.0 means optimization will start on the linear scale at the unmodified 
        exposure profile. Using the log scale for optimization will provide much smoother
        optimization performance because multiplicative steps on the log scale require 
        much less adaptation.

        Parameters
        ----------

        mode : Literal['draws', 'mean', 'manual']
            mode of estimation. mode='mean' takes the mean of the posterior and estimate
            the ECx for this singular value. mode='draws' takes samples from the posterior
            and estimate the ECx for each of the parameter draws. mode='manual' takes
            a parameter set (Dict) in the parameters argument and uses that for estimation. 
            Default: 'draws'
        
        draws : int
            Number of draws to take from the posterior. Only takes effect if mode='draw'.
            Raises an exception if draws < 100, because this is insufficient for a 
            reasonable uncertainty estimate. Default: None (using all samples from the
            posterior)
        
        parameters : Dict[str,float|list[float]]
            a parameter dictionary passed used as model parameters for finding the ECx
            value. Default: None

        log_x0 : float
            the starting value for the multiplication factor of the exposure profile for
            the minimization algorithm. This value is on the log scale. This means, 
            exp(log_x0=0.0) = 1.0, which means that the log_x0=0.0 will start at an
            unmodified exposure profile. Default: 0.0
        
        accept_tol : float
            After optimization is finished, accept_tol is used to assess if the loss
            function for the individual draws exceed a tolerance. These results are
            discarded and a warning is emitted. This is to assert that no faulty optimization
            results enter the estimate. Default: 1e-5
        
        optimizer_tol : float
            Tolerance limit for the minimzer to stop optimization. Default 1e-5

        method : str
            Minization algorithm. See: https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html
            Default: 'cobyla'

        show_plot : bool
            Show the results plot of the lpx. Default: True
        
        force_draws : bool
            Force the estimate method to accept a number of draws less than 100. Default: False

        optimizer_kwargs :
            Additional arguments to pass to the optimizer

        """
        x0_tries = np.array([0.0, -1.0, 1.0, -2.0, 2.0]) + log_x0
        draws = self._check_mode_and_draws_and_parameters(mode, draws, parameters, force_draws)

        self._mode = mode
        mult_factor = []
        loss = []
        iterations = []
        for i in tqdm(range(draws)):
            if mode == "draws":
                sample = self._posterior_sample(i)
            elif mode == "mean":
                sample = self._posterior_mean()
            elif mode == "manual":
                sample = parameters
            else: 
                raise NotImplementedError(
                    f"Bad mode: {mode}. Mode must be one 'mean' or 'draws'"
                )

            success = False
            iteration = 0
            while not success and iteration < len(x0_tries):
                opt_res = minimize(
                    self._loss, x0=x0_tries[iteration], 
                    method=method,
                    tol=optimizer_tol,
                    args=(sample,),
                    **optimizer_kwargs
                ) 

                success = opt_res.fun < accept_tol
                iteration += 1

            # convert to linear scale from log scale
            factor = np.exp(opt_res.x)

            mult_factor.extend(factor)
            iterations.append(iteration)
            loss.append(opt_res.fun)

        res_full = pd.DataFrame(dict(factor = mult_factor, loss=loss, retries=iterations))
        self.results_full = res_full

        metric = "{name}_{x}".format(name=self._name, x=int(self.x * 100),)

        successes = sum(res_full.loss < accept_tol)
        if successes < draws:
            warnings.warn(
                f"Not all optimizations converged on the {metric} ({successes/draws*100}%). " +
                "Adjust starting values and method")
            print(res_full)

        short_msg = f"Estimation success rate: {successes/draws*100}%"
        self.results["msg"] = short_msg
        
        res = res_full.loc[res_full.loss < accept_tol,:]

        if len(res) == 0:
            self.msg = (
                f"{metric} could not be found. Two reasons typically cause this problem: "+
                f"1) no expoure before the time at which the {metric} is calculated. "+
                "Check the the exposure profile. " + 
                f"2) Too high background mortality. If the time at which the {metric} is "+
                f"calculated is large and background mortality is high, the {metric}, " +
                "may be reached independent of the effect and optimization cannot succeed."
            )

            print(self.msg)
            return

        self.results["mean"] = np.round(np.mean(res.factor.values), 4)
        self.results["q05"] = np.round(np.quantile(res.factor.values, 0.05), 4)
        self.results["q95"] = np.round(np.quantile(res.factor.values, 0.95), 4)
        self.results["std"] = np.round(np.std(res.factor.values), 4)
        self.results["cv"] = np.round(np.std(res.factor.values)/np.mean(res.factor.values), 2)

        if show_plot:
            self.plot_profile_and_effect(parameters=parameters)

        print("{name}_{x}".format(name=self._name, x=int(self.x * 100),))
        print(self.results)
        print("\n")

    def plot_profile_and_effect(
        self,
        parameters: Optional[Dict[str,float|List[float]]] = None
    ):
        coordinates_backup = self.sim.coordinates["time"].copy()

        self.sim.coordinates["time"] = np.linspace(0, self.time, 100)
        self.sim.dispatch_constructor()

        if self._mode is None:
            raise GutsBaseError(
                "Run .estimate() before plot_profile_and_effect()"
            )
        elif self._mode == "mean" or self._mode == "draws":
            e_new = self._evaluate(factor=self.results["mean"], theta=self._posterior_mean())
            e_old = self._evaluate(factor=1.0, theta=self._posterior_mean())
        elif self._mode == "manual":
            if parameters is None:
                raise RuntimeError(
                    f"If {self._name}_x was estimated using manual mode, parameters must "+
                    "also be provided here."
                )
            e_new = self._evaluate(factor=self.results["mean"], theta=parameters)
            e_old = self._evaluate(factor=1.0, theta=parameters)
    
        extra_dim = [k for k in list(e_old.results.coords.keys()) if k not in ["time", "id"]]

        if len(extra_dim) > 0:
            labels_old = [
                f"{l} (original)" for l 
                in e_old.results.coords[extra_dim[0]].values
            ]
            labels_new = [
                f"{l} (modified)" for l 
                in e_new.results.coords[extra_dim[0]].values
            ]
        else:
            labels_old = "original"
            labels_new = "modified"



        fig, (ax1, ax2) = plt.subplots(2,1, height_ratios=[1,3], sharex=True)
        ax1.plot(
            e_old.results.time, e_old.results.exposure.isel(id=0), 
            ls="--", label=labels_old,
        )
        ax1.set_prop_cycle(None)
        ax1.plot(
            e_new.results.time, e_new.results.exposure.isel(id=0), 
            label=labels_new
        )


        ax2.plot(
            e_new.results.time, e_new.results.survival.isel(id=0), 
            color="black", ls="--", label="modified"
        )
        ax1.set_prop_cycle(None)

        ax2.plot(
            e_old.results.time, e_old.results.survival.isel(id=0), 
            color="black", ls="-", label="original"
        )
        ax2.hlines(self.x, e_new.results.time[0], self.time, color="grey")
        ax1.set_ylabel("Exposure")
        ax2.set_ylabel("Survival")
        ax2.set_xlabel("Time")
        ax1.legend()
        ax2.legend()
        ax2.set_xlim(0, None)
        ax1.set_ylim(0, None)
        ax2.set_ylim(0, None)
        fig.tight_layout()

        self.figure_profile_and_effect = fig

        self.sim.coordinates["time"] = coordinates_backup
        self.sim.dispatch_constructor()

    

class LPxEstimator(ECxEstimator):
    """
    the LPx is computed, using the existing exposure profile for 
    the specified ID and estimating the multiplication factor for the profile that results
    in an effect of X %
    """
    _name = "LP"

    def __init__(
        self, 
        sim: SimulationBase, 
        id: str,
        x: float=0.5
    ):
        x_in = sim.model_parameters["x_in"].sel(id=[id])
        time = sim.coordinates["time"][-1]
        super().__init__(
            sim=sim, 
            effect="survival", 
            x=x, 
            time=time, 
            x_in=x_in
        )


import pytest

from guts_base import LPxEstimator, GutsBase
from guts_base.sim import construct_sim_from_config, load_idata

@pytest.mark.slow
def test_lp50(sim_post_inference):
    _id = sim_post_inference.observations.id.values[1]
    lpx_estimator = LPxEstimator(sim=sim_post_inference, id=_id)

    theta_mean = lpx_estimator.sim.inferer.idata.posterior.mean(("chain", "draw"))
    theta_mean = {k: v["data"] for k, v in theta_mean.to_dict()["data_vars"].items()}

    lpx_estimator._loss(log_factor=0.0, theta=theta_mean)

    lpx_estimator.plot_loss_curve(mode="draws", draws=2, force_draws=True)
    lpx_estimator.plot_loss_curve(mode="mean")
    lpx_estimator.plot_loss_curve(mode="manual", parameters=lpx_estimator._posterior_mean())

    lpx_estimator.estimate(mode="mean")
    lpx_estimator.estimate(mode="manual", parameters=lpx_estimator._posterior_mean())
    lpx_estimator.estimate(mode="draws", draws=2, force_draws=True)

    lpx_estimator.results
    lpx_estimator.results_full

def test_copy(sim_post_inference):
    """only tests whether the copied estimator can be evaluated"""
    _id = sim_post_inference.observations.id.values[1]
    lpx_estimator = LPxEstimator(sim=sim_post_inference, id=_id)

    e = lpx_estimator.sim.dispatch()
    e()
    e.results


if __name__ == "__main__":
    test_copy(load_idata(construct_sim_from_config("red_sd_ia", GutsBase), "ecx/idata_red_sd_ia.nc"))



import os
import itertools as it
from typing import List
import pandas as pd

from pymob import SimulationBase
from pymob.sim.report import Report, reporting

from guts_base.plot import plot_survival_multipanel
from guts_base.sim.ecx import ECxEstimator

class GutsReport(Report):
    ecx_estimates_times: List = [1, 2, 4, 10]
    ecx_estimates_x: List = [0.1, 0.25, 0.5, 0.75, 0.9]

    def additional_reports(self, sim: "SimulationBase"):
        super().additional_reports(sim=sim)
        self.model_fits(sim)
        self.LCx_estimates(sim)

    @reporting
    def model_fits(self, sim: SimulationBase):
        self._write("### Survival model fits")
        
        out_mp = plot_survival_multipanel(
            sim=sim,
            results=sim.inferer.idata.posterior_model_fits,
            ncols=6,
        )

        lab = self._label.format(placeholder='survival_fits')
        self._write(f"![Surival model fits.\label{{{lab}}}]({os.path.basename(out_mp)})")

        return out_mp


    @reporting
    def LCx_estimates(self, sim):
        X = self.ecx_estimates_x
        T = self.ecx_estimates_times
        P = sim.predefined_scenarios()

        estimates = pd.DataFrame(
            it.product(X, T, P.keys()), 
            columns=["x", "time", "scenario"]
        )

        ecx = []

        for i, row in estimates.iterrows():
            ecx_estimator = ECxEstimator(
                sim=sim,
                effect="survival", 
                x=row.x,
                time=row.time, 
                x_in=P[row.scenario], 
            )
            
            ecx_estimator.estimate(
                mode=sim.ecx_mode,
                draws=250,
                show_plot=False
            )

            ecx.append(ecx_estimator.results)

        results = pd.DataFrame(ecx)
        estimates[results.columns] = results

        out = self._write_table(tab=estimates, label_insert="$LC_x$ estimates")
        
        return out



# TODO: Set up data import pipelines and templates for importing data into expyDB
# TODO: Test plotting
# TODO: Run a simple test case with the ring test data
import pandas as pd
from datetime import datetime, timedelta
import tempfile
from click.testing import CliRunner
from guts_base.data import create_database_and_import_data, OpenGutsIO, time_of_death_to_openguts

def test_data_import_openguts():
    tempdir = tempfile.TemporaryDirectory()

    runner = CliRunner(echo_stdin=True)
    result = runner.invoke(
        create_database_and_import_data, 
        catch_exceptions=False,
        args=[
            "--datasets_path", "data/templates/ringtest_A_SD_openguts_notation.xlsx", 
            "--database_path", f"{tempdir.name}/ringtest.db", 
            "--preprocessing", "guts_base.data.preprocessing.ringtest",
            "--preprocessing-out", f"{tempdir.name}/processed_{{filename}}"
        ]
    )

    if isinstance(result.exception, SystemExit):
        raise KeyError(
            "Invokation of the click command did not execute correctly. " +
            f"Recorded output: {' '.join(result.output.splitlines())}"
        )
    
    else:
        print(result.output)

def test_data_import_time_of_death():
    tempdir = tempfile.TemporaryDirectory()
    runner = CliRunner(echo_stdin=True)
    result = runner.invoke(
        time_of_death_to_openguts, 
        catch_exceptions=False,
        args=[
            "--file", "data/templates/ringtest_A_SD_time_of_death_notation.xlsx", 
            "--sheet", "time-of-death", 
            "--out", tempdir.name,
            "-c", "exposure"
        ]
    )

    if isinstance(result.exception, SystemExit):
        raise KeyError(
            "Invokation of the click command did not execute correctly. " +
            f"Recorded output: {' '.join(result.output.splitlines())}"
        )
    
    else:
        print(result.output)
    
    # this will raise an assertion error if the data cannot be converted to
    # an experiment
    io = OpenGutsIO(f"{tempdir.name}/openguts_ringtest_A_SD_time_of_death_notation.xlsx")
    io.to_experiment()


def test_tod_to_openguts():
    """Tests only the conversion, there is no pre-processing template in place.
    This would have to be built.
    """
    tempdir = tempfile.TemporaryDirectory()
    runner = CliRunner(echo_stdin=True)
    result = runner.invoke(
        time_of_death_to_openguts, 
        catch_exceptions=False,
        args=[
            "--file", "data/templates/tktd_data_template_time_of_death_notation.xlsx", 
            "--sheet", "time-of-death",
            "--out", tempdir.name,
            "-c", "concentration_pesticide__mg/g",
            "-c", "concentration_pesticide_B__mg/g"
        ],
    )

    if isinstance(result.exception, SystemExit):
        raise KeyError(
            "Invokation of the click command did not execute correctly. " +
            f"Recorded output: {' '.join(result.output.splitlines())}"
        )
    
    else:
        print(result.output)
        
    
    processed_data = pd.read_excel(
        f"{tempdir.name}/openguts_tktd_data_template_time_of_death_notation.xlsx",
        sheet_name=None
    )

    pd.testing.assert_frame_equal(
        processed_data["concentration_pesticide__mg_g"],
        pd.DataFrame({
            "time [d]": [0, 1/24, 64],
            "01__1": [0,0,0],
            "02__2": [0,0,0],
        })
    )

    pd.testing.assert_frame_equal(
        processed_data["concentration_pesticide_b__mg_g"],
        pd.DataFrame({
            "time [d]": [0, 1/24, 64],
            "01__1": [0,0,0],
            "02__2": [0.5,0,0],
        })
    )


    pd.testing.assert_frame_equal(
        processed_data["survival"],
        pd.DataFrame({
            "time [d]":  list(range(65)),
            "01__1": [
                30,28,26,26,26,26,26,26,25,24,24,24,24,24,24,24,24,24,24,24,24,
                24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,
                24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,
                24,24],
            "02__2": [
                30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,27,27,27,26,26,25,
                25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,
                25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,
                25,24],
        })
    )

if __name__ == "__main__":
    test_data_import_time_of_death()
    test_tod_to_openguts()


import pathlib
from typing import Dict, Optional, Literal, Protocol, TypedDict, List
import re
import os
import numpy as np
import pandas as pd
import xarray as xr
from pymob.sim.config import Config, DataVariable, Datastructure
from pymob.sim.parameters import Param
from guts_base.sim import GutsBase

__all__ = [
    "PymobSimulator",
]

class ParamsInfoDict(TypedDict):
    name: str
    min: float
    max: float
    initial: float
    vary: bool
    prior: str

class StateVariablesDict(TypedDict):
    dimensions: List[str]
    observed: bool
    y0: List[float]

class Model(Protocol):
    extra_dim: Optional[str]
    params_info: Dict[str, ParamsInfoDict]
    state_variables: Dict[str, StateVariablesDict]
    _params_info_defaults: Dict[str, ParamsInfoDict]
    _it_model: bool

    @staticmethod
    def _rhs_jax():
        raise NotImplementedError
        
    @staticmethod
    def _solver_post_processing():
        raise NotImplementedError

    @staticmethod
    def _likelihood_func_jax():
        raise NotImplementedError


class PymobSimulator(GutsBase):

    @classmethod
    def from_model_and_dataset(
        cls,
        model: Model,
        exposure_data: Dict[str, pd.DataFrame],
        survival_data: pd.DataFrame,
        info_dict: Dict = {},
        pymob_config: Optional[Config] = None,
        output_directory: str|pathlib.Path = pathlib.Path("output/pymob"),
        default_prior: Literal["uniform", "lognorm"] = "lognorm",
    ) -> "PymobSimulator":
        """Construct a PymobSimulator from the 
        """

        if pymob_config is None:
            cfg = Config()
            # Configure: The configuration can be overridden in a subclass to override the 
            # configuration
            cls._configure(config=cfg)
        else:
            cfg = pymob_config

        if isinstance(output_directory, str):
            output_directory = pathlib.Path(output_directory)

        cfg.case_study.output = str(output_directory)

        # overrides scenario path. This means the scenario is also expected in the
        # same folder
        cfg.case_study.scenario_path_override = str(output_directory)
        cfg.case_study.scenario = output_directory.stem
        cfg.case_study.data = cfg.case_study.output_path
        cfg.case_study.observations = "observations.nc"
        cfg.create_directory(directory="results", force=True)
        
        # parse observations
        # obs can be simply subset by selection obs.sel(substance="Exposure-Dime")
        observations = xr.combine_by_coords([
            cls._exposure_data_to_xarray(exposure_data, dim=model.extra_dim),
            cls._survival_data_to_xarray(survival_data)
        ])

        observations.to_netcdf(
            os.path.join(cfg.case_study.output_path, cfg.case_study.observations)
        )

        # configure model and likelihood function
        # extract the fully qualified name of the model module.name
        cfg.simulation.model_class = "{module}.{name}".format(
            module=model.__module__, name=type(model).__name__
        )
        cfg.inference_numpyro.user_defined_error_model = "{module}.{name}".format(
            module=model._likelihood_func_jax.__module__, 
            name=model._likelihood_func_jax.__name__
        )

        # derive data structure and params from the model instance
        cls._set_data_structure(config=cfg, model=model)
        cls._set_params(config=cfg, model=model, default_prior=default_prior)

        # configure starting values and input
        cfg.simulation.x_in = ["exposure=exposure"]
        cfg.simulation.y0 = [f"{k}={v['y0']}" for k, v in model.state_variables.items() if "y0" in v]

        # create a simulation object
        # It is essential that all post processing tasks are done in self.setup()
        # which is extended below. This ensures that the simulation can also be run
        # from automated tools like pymob-infer
        sim = cls(config=cfg)
        sim.setup()
        return sim

    def reset_observations(self):
        """Resets the observations to the original observations after using .from_mempy(...)
        This also resets the sim.coordinates dictionary.
        """

        self.observations = self._obs_backup

    def setup(self, **evaluator_kwargs):
        super().setup(**evaluator_kwargs)
        self._obs_backup = self.observations.copy(deep=True)


    @classmethod
    def _configure(cls, config: Config):
        """This is normally set in the configuration file passed to a SimulationBase class.
        Since the mempy to pymob converter initializes pymob.SimulationBase from scratch
        (without using a config file), the necessary settings have to be specified here.
        """
        config.case_study.output = "results"
        config.case_study.simulation = "PymobSimulator"

        # this must be named guts_base,  whihc is the name of the pip package and
        # this regulates which packages are loaded.
        config.case_study.name = "guts_base"

        config.simulation.x_dimension = "time"
        config.simulation.batch_dimension = "id"
        config.simulation.solver_post_processing = None
        config.simulation.unit_time = "day"
        config.simulation.n_reindexed_x = 100
        config.simulation.forward_interpolate_exposure_data = True
        
        config.inference.extra_vars = ["eps", "survivors_before_t"]
        config.inference.n_predictions = 100

        config.jaxsolver.diffrax_solver = "Tsit5"
        config.jaxsolver.rtol = 1e-10
        config.jaxsolver.atol = 1e-12
        config.jaxsolver.throw_exception = True
        config.jaxsolver.pcoeff = 0.3
        config.jaxsolver.icoeff = 0.3
        config.jaxsolver.dcoeff = 0.0
        config.jaxsolver.max_steps = 1000000
        config.jaxsolver.throw_exception = True


        config.inference_numpyro.gaussian_base_distribution = True
        config.inference_numpyro.kernel = "svi"
        config.inference_numpyro.init_strategy = "init_to_median"
        config.inference_numpyro.svi_iterations = 10_000
        config.inference_numpyro.svi_learning_rate = 0.001

    @classmethod
    def _set_data_structure(cls, config: Config, model: Model):
        """Takes a dictionary that is specified in the model and uses only keys that
        are fields of the DataVariable config-model"""
        
        state_dict = model.state_variables

        config.data_structure = Datastructure(**{
            key: DataVariable(**{
                k: v for k, v in state_info.items()
                if k in DataVariable.model_fields
            })
            for key, state_info in state_dict.items()
        })


    @classmethod
    def _set_params(cls, config: Config, model: Model, default_prior: str):
        params_info = model.params_info

        if model._it_model:
            eps = config.jaxsolver.atol * 10
            params_info["eps"] = {'name':'eps', 'initial':eps, 'vary':False}


        for par, param_dict in params_info.items():
            for k, v in model._params_info_defaults.items():
                if k not in param_dict:
                    param_dict.update({k:v})

        param_df = pd.DataFrame(params_info).T
        param_df["param_index"] = param_df.name.apply(lambda x: re.findall(r"\d+", x))
        param_df["param_index"] = param_df.param_index.apply(lambda x: int(x[0])-1 if len(x) == 1 else None)
        param_df["name"] = param_df.name.apply(lambda x: re.sub(r"\d+", "", x).strip("_"))

        for (param_name, ), group in param_df.groupby(["name"]):

            dims = list(dict.fromkeys(group["dims"]))
            dims = tuple([]) if dims == [None] else tuple(dims)

            prior = list(dict.fromkeys(group["prior"]))
            prior = prior[0] if len(prior) == 1 else prior
            
            _min = np.min(np.ma.masked_invalid(group["min"].values.astype(float)))
            _max = np.max(np.ma.masked_invalid(group["max"].values.astype(float)))
            _init = group["initial"].values.astype(float)
            _free = group["vary"].values

            if isinstance(_min, np.ma.core.MaskedConstant):
                _min = None
            if isinstance(_max, np.ma.core.MaskedConstant):
                _max = None
            
            # TODO: allow for parsing one N-D prior from multiple priors
            # TODO: Another choice would be to parse vary=False priors as deterministic
            #       and use a composite prior from a deterministic and a free prior as
            #       the input into the model

            if prior is None:
                if default_prior == "uniform":
                    _loc = _init * np.logical_not(_free) + _min * _free - config.jaxsolver.atol * 10 * np.logical_not(_free)
                    _scale = _init * np.logical_not(_free) + _max * _free + config.jaxsolver.atol * 10 * np.logical_not(_free)
                    _loc = _loc[0] if len(_loc) == 1 else _loc
                    _scale = _scale[0] if len(_scale) == 1 else _scale
                    prior = f"uniform(loc={_loc},scale={_scale})"
                elif default_prior == "lognorm":
                    _s = 3 * _free + config.jaxsolver.atol * 10 * np.logical_not(_free)
                    _init = _init[0] if len(_init) == 1 else _init
                    _s = _s[0] if len(_s) == 1 else _s

                    prior = f"lognorm(scale={_init},s={_s})"
                else:
                    raise ValueError(
                        f"Default prior: '{default_prior}' is not implemented. "+
                        "Use one of 'uniform', 'lognorm' or specify priors for each "+
                        "parameter directly with: "+
                        f"`model.params_dict['prior'] = {default_prior}(...)`"
                    )

                prior = prior.replace(" ", ",")

            # if isinstance(value,float):
            param = Param(
                value=_init,
                free=np.max(_free),
                min=_min,
                max=_max,
                prior=prior,
                dims=dims
            )

            setattr(config.model_parameters, param_name, param)




import os
import arviz as az
from guts_base.sim import GutsBase

def construct_sim_from_config(
    scenario: str, 
    simulation_class: type, 
    output_path=None
) -> GutsBase:
    """Helper function to construct simulations for debugging"""
    sim = simulation_class(f"scenarios/{scenario}/settings.cfg")

    # this sets a different output directory
    if output_path is not None:
        p = output_path / sim.config.case_study.name / "results" / sim.config.case_study.scenario
        sim.config.case_study.output_path = str(p)
    else:
        sim.config.case_study.scenario = "debug_test"
    sim.setup()
    return sim


def load_idata(sim: GutsBase, idata_file: str) -> GutsBase:
    sim.set_inferer("numpyro")

    if os.path.exists(idata_file):
        sim.inferer.idata = az.from_netcdf(idata_file)
    else:
        sim.inferer.idata = None

    return sim


from typing import List, Optional, Literal
import numpy as np
import xarray as xr
import pandas as pd
import arviz as az
import datetime

import pandas as pd
import numpy as np

from expyDB.intervention_model import (
    Experiment,
    Treatment, 
    Timeseries,
    TsData,
    from_expydb
)


def prepare_dataset(
    idata, 
    variable="survival", 
    unit_time: Literal["day", "hour", "minute", "second"] = "day"
):
    """Get interventions from idata storage with respect ot the treatment 
    ids of the observations and move non indexing-related metadata (unique metadata)
    to the attrs container.
    """
    # this test is guaranteed when prepare dataset is used together with from_expydb
    # because from_expydb organizes the data into datasets with 1 variable,
    # which is the timeseries variable with the coordinates timeseries_id and time
    # for treatments and replicates. Other variables receive their 'own' dataset
    assert len(idata[variable].data_vars) == 1
    array: xr.DataArray = idata[variable][variable]
    array = array.swap_dims(timeseries_id="treatment_id")
    array = array.drop_vars("id")
    # assuming that each timeseries of one variable in each treatment has 
    # a unique name the resulting index should be unique
    array = array.set_index(id=("treatment_id", "timeseries_name"))
    array = array.drop_vars("timeseries_id")
    assert array.indexes["id"].is_unique

    # format time to h and set as float
    time_h = array.time.values / pd.Timedelta(1, unit_time)
    array = array.assign_coords(time=time_h)

    array = move_unique_coordinates_to_attrs(array)

    array.attrs["unit_time"] = unit_time

    # add a unique id for the selected dataset which is only relevant for
    # the scope of modelling
    return array

def move_unique_coordinates_to_attrs(array:xr.DataArray) -> xr.DataArray:
    key: str
    for key, coord in array.coords.items(): # type:ignore
        if key in ["id", "treatment_id", "timeseries_id", "experiment_id", "subject_count", "timeseries_name"]:
            continue
        
        if coord.isnull().all():
            unique_values = [None]
        else:
            unique_values = np.unique(coord.data)

        if len(unique_values) == 1:
            array.attrs.update({key: unique_values[0]})
            array = array.drop_vars(key)
    return array

# def prepare_interventions_dataset(interventions_idata, observations, ivs:Optional[List[str]]=None):
#     """Get interventions from idata storage with respect ot the treatment 
#     ids of the observations"""
#     if ivs is None:
#         ivs = list(interventions_idata.keys())
#     ds_ivs = get_interventions(
#         interventions_idata,
#         observations=observations,
#         ivs=ivs
#     )

#     time_h = ds_ivs.time.values / np.timedelta64(1, "h")
#     ds_ivs = ds_ivs.assign_coords(time=time_h)
#     ds_ivs.attrs["unit_time"] = "hours (h)"

#     return ds_ivs

def to_dataset(
    observations_idata, 
    interventions_idata,
    unit_time: Literal["day", "hour", "minute", "second"] = "hour"
) -> xr.Dataset:
    """Combines intervention and observation datasets, assuming that there is 
    a unique multiindex that can be constructed from 
      - treatment_id
      - timeseries_name

    This way interventions and observations can be combined into a single dataset,
    """
    data_arrays = {}
    for variable in observations_idata.groups():
        # prepare observations
        da = prepare_dataset(
            idata=observations_idata,
            variable=variable,
            unit_time=unit_time,
        )
        data_arrays.update({variable: da})

    # prepare interventions
    for variable in interventions_idata.groups():
        da = prepare_dataset(
            idata=interventions_idata,
            variable=variable,
            unit_time=unit_time,
        )
        data_arrays.update({variable: da})

    return xr.combine_by_coords(data_arrays.values())  # type: ignore


def reduce_multiindex_to_flat_index(dataset):
    multi_index = dataset.id.indexes["id"]
    
    # create a flat index from the multi index
    flat_index = multi_index.map(lambda x: "__".join([str(x_) for x_ in x]))
    
    # remove multi index from dimension 'id'
    dataset = dataset.reset_index("id")
    
    # assign flat index to dimension 'id'
    dataset = dataset.assign_coords(id=flat_index)

    return dataset    

def combine_coords_to_multiindex(
        dataset: xr.Dataset, 
        coordinates: List[str], 
        index_name: str, 
        sep: str = "__"
    ) -> xr.Dataset:
    """Simply combines a list of coordinates into a joint string

    Parameters
    ----------
    dataset : xr.Dataset
        The observations dataset
    coordinates : List[str]
        The coordinates that should be joined
    index_name : str
        The name of the new, joined, coordinate
    sep : str, optional
        The string to separate the coordinate components, by default "__"

    Returns
    -------
    xr.Dataset
        Dataset with a new coordinate composed of the listed coordinates
    """
    try:
        multi_index = pd.MultiIndex.from_arrays([dataset[c].values for c in coordinates])
    except KeyError as err:
        raise KeyError(
            f"Did not find key {err} in the dataset. "
            f"This is probably because the key {err} is equal for all treatments."
        )
    multi_index = multi_index.map(lambda x: "__".join([str(x_) for x_ in x]))
    return dataset.assign_coords({index_name: ("id", multi_index)})

# def get_interventions(interventions_idata, observations, ivs: List[str]) -> xr.Dataset:
#     """Get the interventions according to the treatment ids of the observation
#     dataset.
    
#     Works only for single interventions
#     """
#     X_in = {}
#     for data_var in ivs:
#         x_in = interventions_idata[data_var]\
#             .swap_dims(timeseries_id="treatment_id")\
#             .sel(treatment_id=observations.treatment_id.values)
        
#         x_in = x_in.assign_coords(
#             _id=("treatment_id", range(x_in.sizes["treatment_id"]))
#         )
#         x_in = x_in.swap_dims(treatment_id="_id")
#         X_in.update({data_var: x_in[data_var]})


#     X_in_dataset = xr.concat(X_in.values(), dim="variable")\
#         .assign_coords(variable=ivs)\
#         .to_dataset(dim="variable")
    
#     if "variable" in X_in_dataset.dims:
#         X_in_dataset = X_in_dataset.drop_dims("variable")

#     return X_in_dataset




# def combine_interventions(
#     interventions: az.InferenceData, 
#     force: bool=False
# ) -> xr.DataArray:
#     """Combining interventions into a single dataset is only possible,
#     if there is only a single timeseries for each intervention.

#     Parameters
#     ----------
#     interventions : az.InferenceData
#         Interventions InferenceData. Contains multiple datasets with at
#         least one timeseries
#     force : bool, optional
#         Override restrictions to combine interventions only when the number
#         of timeseries is 1, by default False

#     Returns
#     -------
#     xr.DataArray
#         Interventions, combined into a single dataset

#     Raises
#     ------
#     ValueError
#         If the number of timeseries is larger than 1 and force is not True
#     """
#     assert isinstance(interventions, az.InferenceData)
#     arrays = []
#     for variable, dataset in interventions.items():
#         if dataset.sizes["timeseries_id"] > 1:
#             if force:
#                 arr = dataset.to_array()
#             else:
#                 raise ValueError(
#                     "Combining interventions is only allowed when the number of "
#                     "Timeseries for each variable is 1. This is to avoid blowing "
#                     "Up the size of the dataset with nans, because timeseries ids "
#                     "are different for each variable. You can override this error "
#                     "By using `force=True`"
#                 )
#         else:
#             arr = dataset.squeeze("timeseries_id").to_array()
        
#         arrays.append(arr)

#     return xr.concat(arrays, dim="variable")
    



import numpy as np
import xarray as xr
from scipy.stats import binom
from matplotlib import pyplot as plt
from pymob.utils.testing import assert_no_nans_in_dataset

def prepare_survival_data_for_conditional_binomial(observations: xr.Dataset) -> xr.Dataset:
    """This is a convenience method for preparing survival data for a 
    conditional binomial model. The method simply prepares an array of the
    same size as survival just shifted by one time step to determine the
    number of survivers at the beginning of the next time step to consider
    conditional surviving of repeated observations.
    
    The additional dataset fills NaN values which may occurr in the observations
    but not in the parameters of the distribution by forward filling and
    then fills remaining nans (which can only ocurr in the initial times t)
    with the nominal number of used organisms.
    """
    survival = observations["survival"]
    # fill nan values forward in time with the last observation 
    # until the next observation. Afterwards leading nans are replaced with 
    # the subject count (no lethality observed before the first observation)
    nsurv = survival.ffill(dim="time").fillna(observations.subject_count)

    # Test if the observations that were filled into the dataframe at the beginning
    # are equal to the subject count if available.
    np.testing.assert_array_equal(
        nsurv.isel(time=0, id=~observations.subject_count.isnull()), 
        observations.subject_count.sel(id=~observations.subject_count.isnull())
    )

    assert_no_nans_in_dataset(nsurv.to_dataset())

    # create a convenience observation survivors before t, which gives the
    # number of living organisms at the end of time interval t-1
    # this is used for calculating conditional survival
    observations = observations.assign_coords({
        "survivors_before_t": (("id", "time"), np.column_stack([
            nsurv.isel(time=0).values, 
            nsurv.isel(time=list(range(0, len(nsurv.time)-1))).values
    ]).astype(int))})

    observations = observations.assign_coords({
        "survivors_at_start": (("id", "time"), np.broadcast_to(
            nsurv.isel(time=0).values.reshape(-1,1), 
            shape=nsurv.shape
    ).astype(int))})

    return observations



def survivors_at_start_of_interval(survival: xr.DataArray, ):    
    # create a convenience observation survivors before t, which gives the
    # number of living organisms at the end of time interval t-1
    # this is used for calculating conditional survival
    return np.column_stack([
        survival.isel(time=0).values, 
        survival.isel(time=list(range(0, len(survival.time)-1))).values
    ]).astype(int)


def generate_survival_repeated_observations(
    S, 
    N=10, 
    time=None,
    reps=1, 
    incidence=True,
    seed=1,
    ax=None,
    tol=None
):
    """Generate observations from a survival function S, with N individuals

    For this the conditional survival probability is used. This means that 
    for each time-interval the probability of dying in that interval, conditional
    on having lived until the beginning of that interval.

    S_cond[i] = (S[i-1] - S[i]) / S[i-1] where i are the intervals in T
    
    L[i] = Binom(p=S_cond[i], N=N_alive[i-1])

    L[i] is the death incidence in the interval i. So the number of deceased 
    individuals in the interval.

    For the binomial trials also N changes over time, with 
    
    N_alive[i] = N - sum(L[:i])

    This means the number of alive individuals gets reduced by the cumulative
    number of deceased individuals.

    Parameters
    ----------
    S : ArrayLike
        values from the survival function must be monotonically decreasing
    N : int
        The number of individuals in one experiment that is repeatedly observed
    reps: int
        The number of repeats of the same experiment

    incidence: bool
        If true, returns the number of deaths in each interval. If False returns 
        the number of cumulative deaths until the interval (including the 
        interval).
    """
    rng=np.random.default_rng(seed)

    if time is None:
        time = np.arange(len(S))

    T = len(time)

    if tol is not None: 
        S = np.clip(S, tol, 1-tol)

    L = np.zeros(shape=(reps, T))
    for i in range(T):
        if i == 0: 
            S_0 = 1
        else:
            S_0 = S[i-1]

        # calculate the binomial response of the conditional survival
        # i.e. the probability to die within an interval conditional on 
        # having survived until the beginning of that interval
        L[:, i] = binom(p=(S_0-S[i])/S_0, n=N-L.sum(axis=1).astype(int)).rvs(random_state=rng)


    # observations
    if ax is None:
        fig, ax = plt.subplots(1,1)
    ax.plot(time, S * N, color="black")
    ax.plot(time, N - L.cumsum(axis=1).T, 
            marker="o", color="tab:red", ls="", alpha=.75)
    ax.set_xlabel("Time [h]")
    ax.set_ylabel("Survival")
    ax.set_ylim(N-N*1.02,N*1.02)

    if incidence:
        return L
    else:
        return L.cumsum(axis=1)


from typing import Callable

from . import utils
from . import openguts
from . import expydb
from . import survival
from . import generator
from . import time_of_death
from . import preprocessing

from .openguts import (
    OpenGutsIO,
    create_new_columns_and_test_integrity_of_replicates,
    create_database_and_import_data,
    create_database_and_import_data_main,
    import_data_to_database,
)
from .survival import (
    prepare_survival_data_for_conditional_binomial, 
    survivors_at_start_of_interval,
    generate_survival_repeated_observations
)

from .generator import create_artificial_data, design_exposure_scenario, ExposureDataDict

from .expydb import (
    to_dataset,
    combine_coords_to_multiindex,
    reduce_multiindex_to_flat_index
)

from .time_of_death import (
    time_of_death_to_openguts
)


from guts_base.sim import GutsSimulationVariableExposure, GutsSimulationConstantExposure
from pymob.utils.testing import assert_no_infs_in_dataset, assert_no_nans_in_dataset

def test_guts_constant_exposure():
    sim = GutsSimulationConstantExposure()
    sim.initialize_from_script()

    sim.dispatch_constructor()
    e = sim.dispatch(theta={})
    e()
    e.results

    assert_no_nans_in_dataset(e.results)
    assert_no_infs_in_dataset(e.results)


def test_guts_variable_exposure():
    sim = GutsSimulationVariableExposure()
    sim.initialize_from_script()

    sim.dispatch_constructor()
    e = sim.dispatch(theta={})
    e()
    e.results

    assert_no_nans_in_dataset(e.results)
    assert_no_infs_in_dataset(e.results)




if __name__ == "__main__":
    test_guts_variable_exposure()


"""This module has been developed in the PollinERA project to deal with time of
death notation and add another import format.

TODO: Surpress warnings for too long sheet names
TODO: Skip files that have locked postprocessing files and give a warning
TODO: Give a Status message, which file is being processed
TODO: Apply these changes also to openguts.py
TODO: Write tests for the imports and produce templates. See test_data_import.py
"""


from typing import List
import os
import warnings
from datetime import timedelta
from guts_base.data.utils import datalad_locked_file_warning

import click
import pandas as pd
import numpy as np

# default columns
DEFAULT_COLUMNS_whitespace = dict(
    id_subject = ["id subject", "subject id", "id", "id bee"],
    id_treatment = ["id treatment", "treatment id", "treatment"],
    id_replicate = ["id replicate", "replicate id", "replicate name", "replicate"],
    n = ["individuals", "n", "number_replicates", "n_individuals", "replicates"],
    censored = ["cenzus", "censoring", "escaped"],
    time_start_experiment = ["date of start", "experiment start", "start experiment"],
    time_end_experiment = ["date of end", "experiment end", "end experiment"],
    time_start_exposure = ["time of exposure start", "start exposure", "exposure start"],
    time_end_exposure = ["time of exposure end", "end exposure", "exposure end"],
    time_death = ["time of death", "survival time", "date of death"],
)

DEFAULT_COLUMNS_underscore = {
    k: [v_.replace(" ", "_") for v_ in v] 
    for k, v in DEFAULT_COLUMNS_whitespace.items()
}

DEFAULT_COLUMNS = {
    k: list(set(DEFAULT_COLUMNS_whitespace[k] + DEFAULT_COLUMNS_underscore[k]))
    for k in DEFAULT_COLUMNS_whitespace.keys()
}

REQUIRED_COLUMNS = dict(
    id_subject = True,
    id_treatment = True,
    id_replicate = False,
    censored = False,
    n = False,
    time_start_experiment = True,
    time_end_experiment = False,
    time_start_exposure = True,
    time_end_exposure = True,
    time_death = True,
)



def clean_column_names(columns: List[str]):
    cleaned_columns = []
    for c in columns:
        c = c.lower()  # convert to lowercase
        c = c.strip()  # strip leading and trailing whitespace
        c = c.replace(" ", "_")
        c = c.replace("[", "")
        c = c.replace("]", "")
        c = c.replace("/", "_")

        cleaned_columns.append(c)

    return cleaned_columns

def standardize_column_names(
        columns: List[str], 
        raise_error=True, 
        ignore_columns=[]
    ):
    column_mapper = invert_dict_of_lists(DEFAULT_COLUMNS)
    standardized_columns = []
    for c in columns:
        c = column_mapper.get(c, c) # try to get a standard value for the column

        standardized_columns.append(c)

    missing_columns = [
        k for k in DEFAULT_COLUMNS.keys() 
        if k not in standardized_columns and k not in ignore_columns
    ]
    if len(missing_columns) > 0 and raise_error:
        raise KeyError(
            f"Not all necessary columns could be found. {missing_columns} "
            "could not identified. Rename columns or add the corresponding "
            "columns in the mapper."
        )

    return standardized_columns


def invert_dict_of_lists(original_dict):
    inverted_dict = {}
    for key, value_list in original_dict.items():
        for value in value_list:
            inverted_dict[value] = key
    return inverted_dict

def long_to_wide(df_long, id_columns, time_column, observation_column):
        df_long["id"] = df_long[id_columns].apply(
            lambda x: '__'.join(x.astype(str)), axis=1
        )
        
        df_wide = df_long.pivot(
            # data=df_long.reset_index(), 
            index=time_column,
            values=observation_column, 
            columns="id", 
        )
        
        return df_wide

def wide_to_long(df_wide, id_columns, time_column, observation_column):
    df_long = pd.melt(
        frame=df_wide.reset_index(), 
        value_vars=df_wide.columns,
        id_vars=time_column,
        var_name="id",
        value_name=observation_column
    )

    df_long[id_columns] = df_long.id.str.split("__", n=1, expand=True)
    df_long = df_long.drop(columns="id")
    return df_long[id_columns+[time_column, observation_column]]


def get_unique_value(series, action_if_not_unique="mean"):
    if series.nunique() == 1:
        return series.drop_duplicates().iloc[0]
    else:
        if action_if_not_unique == "mean":
            return series.mean()
        elif action_if_not_unique == "max":
            return series.max()
        elif action_if_not_unique == "min":
            return series.min()
        elif action_if_not_unique == "error":
            raise ValueError("Start time contains non unique values")
        else:
            raise NotImplementedError("Aggregation action is not implemented.")


def make_openguts_intervention_table(
        df: pd.DataFrame, 
        intervention:str,
        intervention_time_unit:str="d",
        rect_interpolate=True,
    ) -> List[pd.DataFrame]:
    # create exposure tables
    id_columns = ["id_treatment", "id_replicate"]
    time_column = f"time [{intervention_time_unit}]"
    
    df_long = []
    for (tid, rid), group in df.groupby(id_columns):
        intervention_value = float(group[intervention].unique())
        if f"time_start_exposure_{intervention}" in group:
            intervention_start = get_unique_value(group[f"time_start_exposure_{intervention}"])
        else:
            intervention_start = get_unique_value(group["time_start_exposure"])
        if f"time_end_exposure_{intervention}" in group:
            intervention_end = get_unique_value(group[f"time_end_exposure_{intervention}"])
        else:
            intervention_end = get_unique_value(group["time_end_exposure"])

        experiment_start = get_unique_value(group["time_start_experiment"])
        experiment_end = get_unique_value(group["time_end_experiment"])

        time = np.array([
            experiment_start,
            intervention_start,
            intervention_end,
            experiment_end
        ])

        value = np.array([0, intervention_value, 0, 0])

        m = pd.DataFrame(
            data=np.column_stack([
                np.repeat(tid, len(time)), 
                np.repeat(rid, len(time)), 
                time, value
            ]), 
            columns=list(group[id_columns].columns) + [time_column, intervention]
        )
        # this throws the first value out if the time of exposure start is
        # identical to the exposure end
        m = m.drop_duplicates(subset=id_columns + [time_column], keep="last")

        df_long.append(m)

    df_long = pd.concat(df_long)
    df_wide =  long_to_wide(df_long, id_columns, time_column, intervention).reset_index()
    df_wide[time_column] = (df_wide[time_column] - experiment_start)
    df_wide = df_wide.set_index(time_column)
    
    if rect_interpolate:
        df_wide = df_wide.reindex(np.unique(np.concatenate([
            np.array(list(df_wide.index)),
            np.array(list(df_wide.index - pd.Timedelta(1, "s")))[1:]
        ])))
        df_wide = df_wide.fillna(method="ffill")
    
    df_wide.index = df_wide.index / pd.Timedelta(1, "d")
    return df_wide



def make_openguts_observation_table(
        df: pd.DataFrame,
        observation="censored",
        observation_schedule:str="d",
    ) -> List[pd.DataFrame]:
    """returns counts of censored individuals"""
    df = df.copy()
    
    experiment_start = get_unique_value(df["time_start_experiment"])
    experiment_end = get_unique_value(df["time_end_experiment"])

    time = pd.date_range(experiment_start, experiment_end, freq=observation_schedule)
    timecol_name = f"time [{observation_schedule.lower()}]"


    id_columns = ["id_treatment", "id_replicate"]

    # calculate survival time
    df[timecol_name] = df["time_death"] - df["time_start_experiment"]

    time_remainder = df[timecol_name] % pd.Timedelta(1, observation_schedule)
    if (time_remainder > pd.Timedelta(0)).any():
        raise ValueError(
            "Observations should be entered at the same time as the experiment start "+
            "df['time_death] - df['time_experiment_start'] should be a multiple of "+
            f"the time resolution of the observation schedule. Here: 1{observation_schedule}"
        )

    if observation == "censored":
        # sum IDs that were marked as censored at time t
        df_long = df.groupby(id_columns+[timecol_name])["censored"].sum()

    elif observation == "lethality":
        # count IDs that died at time t
        df_long = df.groupby(id_columns+[timecol_name])["time_death"].count()

    else:
        raise NotImplementedError(f"observation {observation} is not implemented.")

    df_long = df_long.rename(observation)
    
    # df to wide frame
    df_wide = long_to_wide(df_long.reset_index(), id_columns, timecol_name, observation)
    
    # reindex wide dataframe on time
    df_wide = df_wide.reindex(index=time-experiment_start, method=None)
    df_wide.index = df_wide.index.set_names(timecol_name)
    df_wide = df_wide.fillna(0)

    return df_wide


# write to excel file
def excel_writer(df: pd.DataFrame, file, sheet):
    with warnings.catch_warnings():
        warnings.simplefilter(action="ignore")
        if not os.path.exists(file):
            with pd.ExcelWriter(file, mode="w") as writer:
                df.to_excel(writer, sheet_name=sheet)

        else:
            with pd.ExcelWriter(file, if_sheet_exists="replace", mode="a") as writer:
                df.to_excel(writer, sheet_name=sheet)

def write_data_template(
        notation="time_of_death",
        time_start_experiment=()
    ):
    pass

def time_to_fraction(data, column, experiment_start):
    data


class TimeOfDeathIO:
    def __init__(
        self, 
        file, 
        intervention_columns: List[str],
        sheet:str = "time-of-death",
    ):
        self._file = file
        self.data = self.from_file()

def main(file: str, sheet: str, out:str, intervention_columns: List[str], 
         observation_schedule="d", rect_interpolate=False):
    intervention_columns = clean_column_names(list(intervention_columns))
    processed_file = f"{out}/openguts_{os.path.basename(file)}"
    
    print("\n")
    print(f"Processing File: {file}")
    print(f"Converting from time-of-death to openguts")
    print("-----------------------------------------")

    if os.access(processed_file, os.EX_OK):
        if not os.access(processed_file, os.W_OK):
            datalad_locked_file_warning(processed_file)
            return
    else:
        directory = os.path.dirname(processed_file)
        if os.access(directory, os.EX_OK):
            pass
        else:
            os.makedirs(directory)

    # read datafile
    data = pd.read_excel(io=file, sheet_name=sheet)
    data.columns = clean_column_names(data.columns)

    # Assumptions
    # -----------
    # this should not be too small, Bürger and Focks (2025) assume a topical exposure
    # duration of 1 hour. If exposure duration is too small, it will result in
    # problems with k_d
    exposure_duration = timedelta(seconds=3600)
    exposure_start_delay = timedelta(hours=0)
    id_zfill = 2  # number of zeros to pad ID column values with
    
    # standardize columns
    data.columns = standardize_column_names(data.columns, raise_error=False)
    data["id_treatment"] = data["id_treatment"].astype(str).str.zfill(id_zfill)

    # add optional columns to the dataframe if they are not present
    experiment_start = get_unique_value(data["time_start_experiment"])
    if "time_start_exposure" not in data:
        warnings.warn(
            "No column: 'time_start_exposure'. " 
            f"Assuming time_start_exposure=time_start_experiment"
            f"({experiment_start}) + {exposure_start_delay}",
            category=UserWarning
        )
        exposure_start = experiment_start + exposure_start_delay
        data["time_start_exposure"] = exposure_start
    
    if "time_end_exposure" in data:
        if (data["time_start_experiment"] == data["time_end_exposure"]).all():
            warnings.warn(
                "'time_end_exposure' equals 'time_start_exposure'. "+
                "Removing column 'time_end_exposure'"
            )
            data = data.drop("time_end_exposure", axis=1)
    
    if "time_end_exposure" not in data:
        exposure_start = data["time_start_exposure"]
        warnings.warn(
            "No column: 'time_end_exposure'. " 
            f"Assuming time_end_exposure=time_start_exposure + {exposure_duration}",
            category=UserWarning
        )
        exposure_end = exposure_start + exposure_duration
        data["time_end_exposure"] = exposure_end

    if "time_end_experiment" not in data:
        experiment_end = data.time_death.max()
        warnings.warn(
            "No column: 'time_end_experiment' "
            f"Using the time of the last observation: {experiment_end}",
            category=UserWarning
        )
        data["time_end_experiment"] = experiment_end

    if "id_replicate" not in data:
        warnings.warn(
            "No column: 'id_replicate'. " 
            "Assuming all treatments were only carried out with 1 replicate "
            "(containing n individuals).",
            category=UserWarning
        )
        data["id_replicate"] = 0
        
        # check for replicates
        id_columns = ["id_treatment"]
        for rid, (_, group) in enumerate(data[id_columns+intervention_columns]
                                         .groupby(id_columns)):
            data.loc[group.index, "id_replicate"] = rid + 1

        data["id_replicate"] = data["id_replicate"].astype(str).str.zfill(id_zfill)
        id_columns = ["id_treatment", "id_replicate"]

    elif data["id_replicate"].isna().all():
        warnings.warn(
            "column: 'id_replicate' contained only NAN values" 
            "Assuming all treatments were only carried out with 1 replicate "
            "(containing n individuals).",
            category=UserWarning
        )
        data["id_replicate"] = 0
        
        # check for replicates
        id_columns = ["id_treatment"]
        for rid, (_, group) in enumerate(data[id_columns+intervention_columns]
                                         .groupby(id_columns)):
            data.loc[group.index, "id_replicate"] = rid + 1

        data["id_replicate"] = data["id_replicate"].astype(str).str.zfill(id_zfill)
        id_columns = ["id_treatment", "id_replicate"]
    else:
        data["id_replicate"] = data["id_replicate"].astype(str)
        data["id_treatment"] = data["id_treatment"].astype(str)
        id_columns = ["id_treatment", "id_replicate"]

    if "censored" not in data:
        warnings.warn(
            "No column: 'censoring'. " 
            "Assuming all observation are not censored, meaning each "
            "'time of death' indication comes from an individual that was "
            "observed dead at that time (as opposed to escaped or removed from "
            "the experiment)",
            category=UserWarning
        )
        data["censored"] = 0

    data.columns = standardize_column_names(data.columns, raise_error=True, ignore_columns=["n"])

    (data["time_start_experiment"] - experiment_start).dt.seconds
    (data["time_start_experiment"] - experiment_start).dt.seconds

    interventions = []
    for iv in intervention_columns:
        iv_wide = make_openguts_intervention_table(
            data, 
            intervention=iv, 
            intervention_time_unit="d",
            rect_interpolate=rect_interpolate,
        )
        interventions.append(iv_wide)
        excel_writer(iv_wide, file=processed_file, sheet=iv)

    censored = make_openguts_observation_table(
        data,
        observation="censored",
        observation_schedule=observation_schedule,
    )

    lethality = make_openguts_observation_table(
        data,
        observation="lethality",
        observation_schedule=observation_schedule,
    )

    deaths = lethality - censored

    # excel export
    excel_writer(censored, file=processed_file, sheet="censored")
    excel_writer(lethality, file=processed_file, sheet="lethality (uncensored)")
    excel_writer(deaths, file=processed_file, sheet="lethality (censored)")


    cens_long = wide_to_long(censored, id_columns, f"time [{observation_schedule}]", "censored")
    leth_long = wide_to_long(lethality, id_columns, f"time [{observation_schedule}]", "lethality")
    
    if "n" in data:
        if data["n"].isna().all():
            warnings.warn(
                "column: 'n' contained only NAN values. "+
                "Removed (so it can be created from scratch)",
                category=UserWarning
            )
            data = data.drop("n", axis=1)
        else:
            pass

    if "n" not in data:
        warnings.warn(
            "No column: 'n'. " 
            "Inferring the number of individuals at the beginning of the "
            "experiment from the uncensored number of dead organisms "
            "(including those escaped and alive at the end of the experiment).",
            category=UserWarning
        )
        n = leth_long.groupby(id_columns)["lethality"].sum().rename("n")
        data = pd.merge(data, n.reset_index(), on=id_columns, how="left")


    # calculate survival 
    n = data.groupby(id_columns)["n"].agg("unique").astype(int)
    survival = pd.merge(leth_long, n, on=id_columns, how="left")
    mortality = survival.groupby(id_columns)["lethality"].cumsum()
    survival["survival"] = survival["n"] - mortality
    survival_wide = long_to_wide(survival, id_columns, f"time [{observation_schedule}]", "survival")
    excel_writer(survival_wide, file=processed_file, sheet="survival")
    
    # Calculate the number of present organisms just after censoring
    # n_observed_after_censoring = survival_wide.copy()
    # n_observed_after_censoring[survival_wide.columns] = np.row_stack([
    #     survival_wide.iloc[0].values,
    #     survival_wide.iloc[:-1].values - censored.iloc[1:].values
    # ])
    # excel_writer(n_observed_after_censoring, file=processed_file, 
    #              sheet="n_observed_after_censoring")


    # Calculate the number of organisms alive after the last observation
    n_observed_after_last_observation = survival_wide.copy()
    n_observed_after_last_observation[survival_wide.columns] = np.row_stack([
        np.full_like(survival_wide.iloc[0].values, np.nan),
        survival_wide.iloc[:-1].values
    ])
    excel_writer(n_observed_after_last_observation, file=processed_file, 
                 sheet="n_observed_after_last_observation")


    data.columns = standardize_column_names(data.columns)
    data_minimal = data[list(DEFAULT_COLUMNS.keys()) + intervention_columns]
    excel_writer(data_minimal.set_index("id_subject"), file=processed_file, 
                 sheet="time_of_death")

    if "meta" in pd.ExcelFile(file).sheet_names:
        excel_writer(
            df=pd.read_excel(file, sheet_name="meta").set_index("Metadata"),
            file=processed_file,
            sheet="meta"
        )
    elif "Info" in pd.ExcelFile(file).sheet_names:
        metadata = pd.read_excel(io=file, sheet_name="Info")
        metadata = metadata.set_index("Experiment information")
        metadata.columns = ["value", "description"]
        metadata.loc["interventions", "value"] = ", ".join(intervention_columns)
        metadata.loc["observations", "value"] = ", ".join(["survival", "censored"])
        excel_writer(metadata, file=processed_file, sheet="meta")
    else:
        warnings.warn("No metadata found in sheets 'meta' or 'Info'.")


@click.command()
@click.option("--file", "-f", help="Path to the xlsx file")
@click.option("--sheet", "-s", help="Name of the excel sheet")
@click.option("--out", "-o", help="Output directory", default="processed_data")
@click.option("--observation_schedule", help="Schedule of the observations: d - daily, h - hourly", default="d")
@click.option("--intervention_columns", "-c", multiple=True, type=str, help="Names of the columns that carry the exposure information")
def time_of_death_to_openguts(file, sheet, out, observation_schedule, intervention_columns):
    main(
        file=file, 
        sheet=sheet, 
        out=out, 
        intervention_columns=intervention_columns,
        observation_schedule=observation_schedule
    )


if __name__ == "__main__":


    if os.path.basename(os.getcwd()) != "data":
        os.chdir("case_studies/tktd-osmia/data")
        # call the underlying function
        ctx = click.Context(time_of_death_to_openguts)
        ctx.forward(
            time_of_death_to_openguts,
            file="test/template_time_of_death.xlsx",
            sheet="time-of-death",
            intervention_columns=["Substance_A", "Substance B"],
        )
    else:
        time_of_death_to_openguts()



import numpy as np
import xarray as xr
from typing import TypedDict, Dict, Optional, Sequence

class ExposureDataDict(TypedDict):
    start: int
    end: int
    exposure: float|Sequence[float]

def create_artificial_data(
    t_max, 
    dt, 
    exposure_paths=["oral", "topical", "contact"],
    intensity=[0.1, 0.5, 0.05],
    seed=1,
):
    rng = np.random.default_rng(1)
    time = np.arange(0, t_max, step=dt)  # daily time resolution

    # calculate potential exposure based on a lognormal distribution
    oral = rng.lognormal(mean=np.log(intensity[0]), sigma=0.5, size=len(time))
    # and include a random exposure days
    oral *= rng.binomial(n=1, p=1, size=len(time))


    # calculate potential exposure based on a lognormal distribution
    topical = rng.lognormal(mean=np.log(intensity[1]), sigma=1, size=len(time))
    # and include a random exposure days
    topical *= rng.binomial(n=1, p=0.25, size=len(time))


    # calculate potential exposure based on a lognormal distribution
    contact = rng.lognormal(mean=np.log(intensity[2]), sigma=0.1, size=len(time))
    # and include a random exposure days
    contact *= rng.binomial(n=1, p=0.8, size=len(time))



    exposures = xr.Dataset(
        data_vars={
            "exposure": (("time", "exposure_path"), np.column_stack([oral, topical, contact])),
        },
        coords={"time": time, "exposure_path": ["oral", "topical", "contact"]}
    )

    return exposures.sel(exposure_path=exposure_paths)


def design_exposure_timeseries(time: Sequence[float], exposure: ExposureDataDict, eps: float):
    if exposure is None:
        return
    
    exposure["end"] = time[-1] if exposure["end"] is None else exposure["end"]

    return np.where(
        np.logical_and(time >= exposure["start"], time < exposure["end"]),
        exposure["concentration"],
        0
    )

def design_exposure_scenario(
    t_max: float, 
    dt: float, 
    exposures: Dict[str,ExposureDataDict],
    eps: float = 1e-8,
    exposure_dimension: str = "exposure_type",
):
    """
    TODO: tmax, dt and eps are probably not necessary
    """
    # add dt so that tmax is definitely inclded
    time = np.arange(0, t_max+dt, step=dt)  # daily time resolution
    time = np.unique(np.concatenate([time] + [
        np.array([time[-1] if vals["end"] is None else vals["end"]])
        for key, vals in exposures.items()

    ]))

    treatments = {}
    for key, expo in exposures.items():
        treat = design_exposure_timeseries(time, expo, eps)
        treatments.update({key: treat})

    data = np.column_stack(list(treatments.values()))
    data = np.expand_dims(data, axis=0)

    coords = {"id": [0], "time": time}
    
    coords.update({exposure_dimension: list(treatments.keys())})

    exposures_dataset = xr.Dataset(
        data_vars={"exposure": (tuple(coords.keys()), data)},
        coords=coords
    )

    return exposures_dataset




import numpy as np
import arviz as az
from matplotlib import pyplot as plt

from pymob.sim.plot import SimulationPlot
from pymob import SimulationBase

def plot_prior_predictions(
    sim: SimulationBase, 
    data_vars=["survival"],
    title_func=lambda sp, c: f"{sp.observations.id.values[0]}"
):
    idata = sim.inferer.prior_predictions() # type: ignore
    def plot_survival_data_probs(idata):
        return idata["survival"] * sim.observations.sel(id=[id_], time=0).survival
    
    fig, axes = plt.subplots(
        ncols=sim.observations.dims["id"], nrows=len(data_vars), 
        sharex=True, sharey="row", figsize=(30,10), squeeze=False)
    for i, id_ in enumerate(sim.coordinates["id"]):
        simplot = SimulationPlot(
            observations=sim.observations.sel(id=[id_]),
            idata=idata.sel(id=[id_]),  # type: ignore
            rows=data_vars,
            coordinates=sim.coordinates,
            config=sim.config,
            obs_idata_map={
                "survival": plot_survival_data_probs
            },
            idata_groups=["prior_model_fits"],  # type: ignore
        )
        # replace simplot axis
        for j, k in enumerate(simplot.rows):
            simplot.axes_map[k]["all"] = axes[j][i]

        simplot.plot_data_variables()
        simplot.set_titles(title_func)
        for j, k in enumerate(simplot.rows):
            if i != 0:
                simplot.axes_map[k]["all"].set_ylabel("")
            if j != 3:
                simplot.axes_map[k]["all"].set_xlabel("")

        simplot.close()
    
    fig.tight_layout()
    fig.savefig(f"{sim.output_path}/combined_prior_predictions.png")


def plot_posterior_predictions(
    sim: SimulationBase, 
    data_vars=["survival"],
    title_func=lambda sp, c: f"{sp.observations.id.values[0]}",
    groups=["posterior_model_fits", "posterior_predictive"],
):
    fig, axes = plt.subplots(
        ncols=sim.observations.dims["id"], nrows=len(data_vars), 
        sharex=True, sharey="row", figsize=(30,10), squeeze=False
    )

    def plot_survival_data_probs_and_preds(dataset):
        if dataset.attrs["group"] == "posterior_model_fits":
            return dataset["survival"] * sim.observations.sel(id=[id_], time=0).survival
        if dataset.attrs["group"] == "posterior_predictive":
            return dataset["survival"]

    for i, id_ in enumerate(sim.coordinates["id"]):
        simplot = SimulationPlot(
            observations=sim.observations.sel(id=[id_]),
            idata=sim.inferer.idata.sel(id=[id_]),  # type: ignore
            rows=data_vars,
            coordinates=sim.coordinates,
            config=sim.config,
            obs_idata_map={
                "survival": plot_survival_data_probs_and_preds,
            },
            idata_groups=groups,  # type: ignore
        )
        
        # replace simplot axis
        for j, k in enumerate(simplot.rows):
            simplot.axes_map[k]["all"] = axes[j][i]

        simplot.plot_data_variables()
        simplot.set_titles(title_func)
        for j, k in enumerate(simplot.rows):
            if i != 0:
                simplot.axes_map[k]["all"].set_ylabel("")
            if j != 3:
                simplot.axes_map[k]["all"].set_xlabel("")

        simplot.close()
    
    fig.tight_layout()
    fig.savefig(f"{sim.output_path}/combined_posterior_predictions.png")


def plot_survival(sim, results):
    fig, ax = plt.subplots(1,1)
    obs = sim.observations.survival / sim.observations.subject_count
    ax.plot(sim.observations.time, obs.T, marker="o", color="black")
    ax.plot(results.time, results.survival.T, color="black")


def plot_survival_multipanel(sim, results, ncols=6, title=lambda _id: _id):

    n_panels = results.sizes["id"]

    nrows = int(np.ceil(n_panels / ncols))

    fig, axes = plt.subplots(nrows, ncols, sharex=True, sharey=True, figsize=(ncols*2+2, nrows*1.5+2))
    axes = axes.flatten()
    mean = results.mean(("chain", "draw"))
    hdi = az.hdi(results, 0.95)
    survival = sim.observations.survival / sim.observations.survival.isel(time=0)

    plot_kwargs = {"color": "black"}
    # param_cycler = plt.rcParams['axes.prop_cycle']
    for _id, ax in zip(sim.observations.id.values, axes):
        ax.set_ylim(-0.05,1.05)

        # TODO: use time unit from observations (?)
        ax.set_xlabel("Time")
        ax.set_ylabel("Survival")
        ax.plot(mean.time, mean.sel(id=_id).survival.T, **plot_kwargs)
        ax.fill_between(hdi.time, *hdi.sel(id=_id).survival.T, alpha=.5, **plot_kwargs)
        ax.plot(survival.time, survival.sel(id=_id).T, ls="", marker="o", alpha=.5, **plot_kwargs)
        ax.set_title(title(_id))

    out = f"{sim.output_path}/survival_multipanel.png"
    fig.tight_layout()
    fig.savefig(out)

    return out


def multipanel_title(sim, _id): 
    oid = sim.observations.sel(id=_id)
    exposure_path = oid.exposure_path.values
    rac = np.round(oid.concentration_closer_x_rac.max().values  * 100, 3)
    return "{ep}\n{c} %RAC".format(ep=str(exposure_path), c=str(rac))

def plot_intermediate_results(sim: SimulationBase, id=0):
    e = sim.dispatch()
    e()
    e.results

    results = e.results.isel(id=[id])

    datavars = sim.config.data_structure.data_variables
    fig, axes = plt.subplots(len(datavars), 1)

    for ax, dv in zip(axes, datavars):
        res = results[dv]
        if len(res.shape) > 2:
            res_ = res.transpose(..., "time", sim.config.simulation.batch_dimension)
            for r in res_:
                ax.plot(r.time, r, alpha=.5)
        else:
            ax.plot(results.time, res.transpose("time", sim.config.simulation.batch_dimension), alpha=.5)

        ax.set_ylabel(dv)



from . import base
from . import ecx
from . import report
from . import utils

from .base import (
    GutsBase,
    GutsSimulationConstantExposure,
    GutsSimulationVariableExposure
)

from .ecx import ECxEstimator, LPxEstimator
from .report import GutsReport

from .mempy import PymobSimulator
from .utils import (
    GutsBaseError
)

from .constructors import (
    construct_sim_from_config, 
    load_idata
)


import os

import pandas as pd

from expyDB.intervention_model import (
    Experiment, Treatment, Timeseries, 
    PandasConverter,
)

def read_timeseries_sheet(path, sheet, sep=None):
    ts = pd.read_excel(path, sheet_name=sheet, index_col=0)  # type: ignore
    multi_index = pd.MultiIndex.from_tuples(
        [tuple(c.split(sep)) for c in ts.columns], names=["treatment_id", "timeseries_id"]
    )
    ts.columns = multi_index
    return ts

def ringtest(path, new_path):
    exposure = read_timeseries_sheet(path, sheet="Exposure", sep=" ")
    exposure.index.name = "time"
    survival = read_timeseries_sheet(path, sheet="Survival", sep=" ")
    survival.index.name = "time"

    # TODO: possibly using a normal index would also be acceptable
    template = PandasConverter(Experiment())
    # template.meta.index = template.meta_multiindex

    # extract information from the meta that is needed elsewhere
    data = {}
    data.update({"exposure": exposure})
    data.update({"survival": survival})


    map = [
        # new keys
        (None, ("experiment", "name"), lambda x: "Ring test"),
        (None, ("experiment", "interventions"), lambda x: ["exposure"]),
        (None, ("experiment", "observations"), lambda x: ["survival"]),
        (None, ("experiment", "public"), lambda x: True),

        (None, ("treatment", "medium"), lambda x: "water"),

        (None, ("observation", "unit"), lambda x: "-"),
        (None, ("observation", "time_unit"), lambda x: "day"),

        (None, ("intervention", "unit"), lambda x: "-"),
        (None, ("intervention", "time_unit"), lambda x: "day"),
    ]

    template.map_to_meta(map=map)
    template.data = data
    template.to_excel(new_path)

    return new_path




import os
import glob
from functools import partial
from copy import deepcopy
import importlib
import warnings
import numpy as np
import xarray as xr
from diffrax import Dopri5
from typing import Literal, Optional, List, Dict, Mapping, Sequence, Tuple
import tempfile
import pandas as pd

from pymob import SimulationBase
from pymob.sim.config import DataVariable, Param, string_to_list, NumericArray

from pymob.solvers import JaxSolver
from pymob.solvers.base import rect_interpolation
from expyDB.intervention_model import (
    Treatment, Timeseries, select, from_expydb
)


from guts_base.sim.utils import GutsBaseError
from guts_base import mod
from guts_base.data import (
    to_dataset, reduce_multiindex_to_flat_index, create_artificial_data, 
    create_database_and_import_data_main, design_exposure_scenario, ExposureDataDict
)
from guts_base.sim.report import GutsReport

class GutsBase(SimulationBase):
    """
    Initializes GUTS models from a variety of data sources 

    Initialization follows a couple of steps
    1. check if necessary entries are made in the configuration, otherwise add defaults
    2. read data or take from input
    3. process data (add dimensions, or add indices)
    4. Prepare model input
    """
    solver = JaxSolver
    Report = GutsReport
    results_interpolation: Tuple[float,float,int] = (np.nan, np.nan, 100)
    _skip_data_processing: bool = False

    def initialize(self, input: Optional[Dict] = None):
        """Initiaization goes through a couple of steps:

        1. Configuration: This makes case-study specific changes to the configuration
            file or sets state variables that are relevant for the simulation
            TODO: Ideally everything that is configurable ends up in the config so it
            can be serialized

        2. Import data: This method consists of submethods that can be adapted or 
            overwritten in subclass methods.
            - .read_data
            - .save_observations
            - .process_data
            process_data itself utilizes the submethods _create_indices and 
            _indices_to_dimensions which are empty methods by default, but can be used
            in subclasses if needed

        3. Initialize the simulation input (parameters, y0, x_in). This can 

        By splitting up the simulation init method, into these three steps, modifcations
        of the initialize method allows for higher granularity in subclasses.
        """

        # 1. Configuration
        self.configure_case_study()

        # 2. Import data
        self.observations = self.read_data()
        # FIXME: Saving observations here is not intuituve. If i export a simulation,
        # I want to use the last used state, not some obscure intermediate state
        # self.save_observations(filename="observations.nc", directory=self.output_path, force=True)
        if not self._skip_data_processing:
            self.process_data()

        # 3. prepare y0 and x_in
        self.prepare_simulation_input()

    def configure_case_study(self):
        """Modify configuration file or set state variables
        TODO: This should only modify the configuration file, so that changes
        are transparent.
        """
        if self._model_class is not None:
            self.model = self._model_class._rhs_jax
            self.solver_post_processing = self._model_class._solver_post_processing

        self.ecx_mode: Literal["mean", "draws"] = "mean"

        self.unit_time: Literal["day", "hour", "minute", "second"] = "day"
        if hasattr(self.config.simulation, "unit_time"):
            self.unit_time = self.config.simulation.unit_time  # type: ignore

        if hasattr(self.config.simulation, "skip_data_processing"):
            self._skip_data_processing = bool(self.config.simulation.skip_data_processing)  # type: ignore

        if hasattr(self.config.simulation, "results_interpolation"):
            results_interpolation_string = string_to_list(self.config.simulation.results_interpolation)
            self.results_interpolation = (
                float(results_interpolation_string[0]),
                float(results_interpolation_string[1]),
                int(results_interpolation_string[2])
            )

    def prepare_simulation_input(self):
        x_in = self.parse_input(input="x_in", reference_data=self.observations, drop_dims=[])
        y0 = self.parse_input(input="y0", reference_data=self.observations, drop_dims=["time"])
        
        # add model components
        if self.config.simulation.forward_interpolate_exposure_data: # type: ignore
            self.model_parameters["x_in"] = rect_interpolation(x_in)
        else:
            self.model_parameters["x_in"] = x_in

        self.model_parameters["y0"] = y0
        self.model_parameters["parameters"] = self.config.model_parameters.value_dict

    def construct_database_statement_from_config(self):
        """returns a statement to be used on a database"""
        substance = self.config.simulation.substance # type:ignore
        exposure_path = self.config.simulation.exposure_path # type:ignore
        return (
            select(Timeseries, Treatment)
            .join(Timeseries)
        ).where(
            Timeseries.variable.in_([substance]),  # type: ignore
            Timeseries.name == {exposure_path}
        )

    def read_data(self):
        # TODO: Update to new INTERVENTION MODEL
        dataset = str(self.config.case_study.observations)
        
        # read from a directory
        if os.path.isdir(os.path.join(self.config.case_study.data_path, dataset)):
            # This looks for xlsx files in the folder and imports them as a database and
            # then proceeds as normal
            files = glob.glob(os.path.join(
                self.config.case_study.data_path, 
                dataset, "*.xlsx"
            ))

            tempdir = tempfile.TemporaryDirectory()
            dataset = self.read_data_from_xlsx(data=files, tempdir=tempdir)

        ext = dataset.split(".")[-1]
        
        if not os.path.exists(dataset):
            dataset = os.path.join(self.data_path, dataset)
            
        if ext == "db":
            statement = self.construct_database_statement_from_config()
            observations = self.read_data_from_expydb(dataset, statement)
            
            # TODO: Integrate interventions in observations dataset

        elif ext == "nc":
            observations = xr.load_dataset(dataset)

        else:
            raise NotImplementedError(
                f"Dataset extension '.{ext}' is not recognized. "+
                "Please use one of '.db' (mysql), '.nc' (netcdf)."
            )
        
        return observations
        
    def read_data_from_xlsx(self, data, tempdir):
        database = os.path.join(tempdir.name, "import.db")

        if hasattr(self.config.simulation, "data_preprocessing"):
            preprocessing = self.config.simulation.data_preprocessing
        else:
            preprocessing = None

        create_database_and_import_data_main(
            datasets_path=data, 
            database_path=database, 
            preprocessing=preprocessing,
            preprocessing_out=os.path.join(tempdir.name, "processed_{filename}")
        )

        return database    


    def read_data_from_expydb(self, database, statement) -> xr.Dataset:

        observations_idata, interventions_idata = from_expydb(
            database=f"sqlite:///{database}",
            statement=statement
        )

        dataset = to_dataset(
            observations_idata, 
            interventions_idata,
            unit_time=self.unit_time
        )
        dataset = reduce_multiindex_to_flat_index(dataset)

        # "Continue here. I want to return multidimensional datasets for data coming "+
        # "from the database. The method can be implemented in any class. Currently I'm looking "+
        # "at guts base"

        filtered_dataset = self.filter_dataset(dataset)

        return filtered_dataset

    def process_data(self):
        """
        Currently these methods, change datasets, indices, etc. in-place.
        This is convenient, but more difficult to re-arragen with other methods
        TODO: Make these methods static if possible

        """
        self._create_indices()
        self._indices_to_dimensions()

        # define tolerance based on the sovler tolerance
        self.observations = self.observations.assign_coords(eps=self.config.jaxsolver.atol * 10)

        self._reindex_time_dim()

        if "survival" in self.observations:
            if "subject_count" not in self.observations.coords:
                self.observations = self.observations.assign_coords(
                    subject_count=("id", self.observations["survival"].isel(time=0).values, )
                )
            self.observations = self._data.prepare_survival_data_for_conditional_binomial(
                observations=self.observations
            )

        if "exposure" in self.observations:
            self.config.data_structure.exposure.observed=False

    def _create_indices(self):
        """Use if indices should be added to sim.indices and sim.observations"""
        pass

    def _indices_to_dimensions(self):
        pass

    def filter_dataset(self, dataset: xr.Dataset) -> xr.Dataset:
        return dataset

    def _reindex_time_dim(self):
        if self.config.simulation.model is not None:
            if "_it" in self.config.simulation.model.lower():
                self.logger.info(msg=(
                    "Redindexing time vector to increase resolution, because model has "+
                    "'_it' (individual tolerance) in it's name"
                ))
                if not hasattr(self.config.simulation, "n_reindexed_x"): 
                    self.config.simulation.n_reindexed_x = 100

                new_time_index = np.unique(np.concatenate([
                    self.coordinates["time"],
                    np.linspace(
                        0, np.max(self.coordinates["time"]), 
                        int(self.config.simulation.n_reindexed_x) # type: ignore
                    )
                ]))
                self.observations = self.observations.reindex(time = new_time_index)
                return

        self.logger.info(msg=(
            "No redindexing of time vector to, because model name did not contain "+
            "'_it' (individual tolerance), or model was not given by name. If an IT model "
            "is calculated without a dense time resolution, the estimates can be biased!"
        ))

    def recompute_posterior(self):
        """This function interpolates the posterior with a given resolution
        posterior_predictions calculate proper survival predictions for the
        posterior.

        It also makes sure that the new interpolation does not include fewer values
        than the original dataset
        """

        ri = self.results_interpolation

        # generate high resolution posterior predictions
        if self.results_interpolation is not None:
            time_interpolate = np.linspace(
                start=float(self.observations["time"].min()) if np.isnan(ri[0]) else ri[0],
                stop=float(self.observations["time"].max()) if np.isnan(ri[0]) else ri[1],
                num=self.results_interpolation[2] 
            )

            # combine original coordinates and interpolation. This 
            # a) helps error checking during posterior predictions.
            # b) makes sure that the original time vector is retained, which may be
            #    relevant for the simulation success (e.g. IT model)
            obs = self.observations.reindex(
                time=np.unique(np.concatenate(
                    [time_interpolate, self.observations["time"]]
                )),
            )

            obs["survivors_before_t"] = obs.survivors_before_t.ffill(dim="time").astype(int)
            obs["survivors_at_start"] = obs.survivors_at_start.ffill(dim="time").astype(int)
            self.observations = obs
            
        self.dispatch_constructor()
        _ = self._prob.posterior_predictions(self, self.inferer.idata) # type: ignore
        self.inferer.store_results(output=f"{self.output_path}/numpyro_posterior_interp.nc") # type: ignore
        self.logger.info("Recomputed posterior and storing in `numpyro_posterior_interp.nc`")


    def prior_predictive_checks(self, **plot_kwargs):
        super().prior_predictive_checks(**plot_kwargs)

        self._plot.plot_prior_predictions(self, data_vars=["survival"])

    def posterior_predictive_checks(self, **plot_kwargs):
        super().posterior_predictive_checks(**plot_kwargs)

        self.recompute_posterior()
        # TODO: Include posterior_predictive group once the survival predictions are correctly working
        self._plot.plot_posterior_predictions(self, data_vars=["survival"], groups=["posterior_model_fits"])


    def plot(self, results):
        self._plot.plot_survival(self, results)

    def predefined_scenarios(self):
        """
        TODO: Fix timescale to observations
        TODO: Incorporate extra exposure patterns (constant, pulse_1day, pulse_2day)
        """
        # get the maximum possible time to provide exposure scenarios that are definitely
        # long enough
        time_max = max(
            self.observations[self.config.simulation.x_dimension].max(), 
            *self.Report.ecx_estimates_times
        )

        # this produces a exposure x_in dataset with only the dimensions ID and TIME
        standard_dimensions = (
            self.config.simulation.batch_dimension,
            self.config.simulation.x_dimension, 
        )

        # get dimensions different from standard dimensions
        exposure_dimension = [
            d for d in self.observations.exposure.dims if d not in 
            standard_dimensions
        ]

        # raise an error if the number of extra dimensions is larger than 1
        if len(exposure_dimension) > 1:
            raise ValueError(
                f"{type(self).__name__} can currently handle one additional dimension for "+
                f"the exposure beside {standard_dimensions}. You provided an exposure "+ 
                f"array with the dimensions: {self.observations.exposure.dims}"
            )
        else:
            exposure_dimension = exposure_dimension[0]

        # iterate over the coordinates of the exposure dimensions to 
        exposure_coordinates = self.observations.exposure[exposure_dimension].values

        scenarios = {}
        for coord in exposure_coordinates:
            concentrations = np.where(coord == exposure_coordinates, 1.0, 0.0)

            exposure_dict = {
                coord: ExposureDataDict(start=0, end=1, concentration=conc)
                for coord, conc in zip(exposure_coordinates, concentrations)
            }

            scenario = design_exposure_scenario(
                exposures=exposure_dict,
                t_max=time_max,
                dt=1/24,
                exposure_dimension=exposure_dimension
            )

            scenarios.update({
                f"1day_exposure_{coord}": scenario
            })

        return scenarios

    @staticmethod
    def _exposure_data_to_xarray(exposure_data: Dict[str, pd.DataFrame], dim: str):
        """
        TODO: Currently no rect interpolation
        """
        arrays = {}
        for key, df in exposure_data.items():
            # this override is necessary to make all dimensions work out
            df.index.name = "time"
            arrays.update({
                key: df.to_xarray().to_dataarray(dim="id", name=key)
            }) 

        exposure_array = xr.Dataset(arrays).to_array(dim=dim, name="exposure")
        exposure_array = exposure_array.transpose("id", "time", ...)
        return xr.Dataset({"exposure": exposure_array})

    @staticmethod
    def _survival_data_to_xarray(survival_data: pd.DataFrame):
        # TODO: survival name is currently not kept because the raw data is not transferred from the survival
        survival_data.index.name = "time"
        
        survival_array = survival_data.to_xarray().to_dataarray(dim="id", name="survival")
        survival_array = survival_array.transpose("id", "time", ...)
        arrays = {"survival": survival_array}
        return xr.Dataset(arrays)


    def expand_batch_like_coordinate_to_new_dimension(self, coordinate, variables):
        """This method will take an existing coordinate of a dataset that has the same
        coordinate has the batch dimension. It will then re-express the coordinate as a
        separate dimension for the given variables, by duplicating the N-Dimensional array
        times the amount of unique names in the specified coordinate to create an 
        N+1-dimensional array. This array will be filled with zeros along the batch dimension
        where the specified coordinate along the ID dimension coincides with the new (unique)
        coordinate of the new dimension. 

        This process is entirely reversible
        """
        old_coords = self.observations[coordinate]
        batch_dim = self.config.simulation.batch_dimension

        # old coordinate before turning it into a dimension
        obs = self.observations.drop(coordinate)

        # create unique coordinates of the new dimension, preserving the order of the
        # old coordinate
        _, index = np.unique(old_coords, return_index=True)
        coords_new_dim = tuple(np.array(old_coords)[sorted(index)])

        for v in variables:
            # take data variable and extract dimension order
            data_var = obs[v]
            dim_order = data_var.dims

            # expand the dimensionality, then transpose for new dim to be last
            data_var = data_var.expand_dims(coordinate).transpose(..., batch_dim, coordinate)

            # create a dummy dimension to broadcast the new array 
            # dummy_3d = np.ones((1, len(coords_new_dim)))
            dummy_categorical = pd.get_dummies(old_coords).astype(int).values

            # apply automatic broadcasting to increase the size of the new dimension
            # data_var_np1_d = data_var * dummy_3d
            data_var_np1_d = data_var * dummy_categorical

            # annotate coordinates of the new dimension
            data_var_np1_d = data_var_np1_d.assign_coords({
                coordinate: list(coords_new_dim)
            })

            # transpose back to original dimension order with new dim as last dim
            data_var_np1_d = data_var_np1_d.transpose(*dim_order, coordinate)
            obs[v] = data_var_np1_d

        return obs


    def reduce_dimension_to_batch_like_coordinate(self, dimension, variables):
        """This method takes an existing dimension from a N-D array and reduces it to an
        (N-1)-D array, by writing a new coordinate from the reducible dimension in the way
        that the new batch-like coordinate takes the coordinate of the dimension, where
        the data of the N-D array was not zero. After it has been asserted that there is
        only a unique candidate for the each coordinate along the batch dimension 
        (i.e. only one value is non-zero for a given batch-coordinate), the dimension will
        be reduced by summing over the given dimension.

        The method is contingent on having no overlap in batch dimension in the dataset
        """
        pass

    def initialize_from_script(self):
        pass

    @property
    def _model_class(self):
        if hasattr(self.config.simulation, "model_class"):
            module, attr = self.config.simulation.model_class.rsplit(".", 1)
            _module = importlib.import_module(module)
            return getattr(_module, attr)
        else:
            return None

    ### API methods ###


    def point_estimate(
        self, 
        estimate: Literal["mean", "map"] = "map",  
        to: Literal["xarray", "dict"] = "xarray"
    ):
        """Returns a point estimate of the posterior. If you want more control over the posterior
        use the attribute: sim.inferer.idata.posterior and summarize it or select from it
        using the arviz (https://python.arviz.org/en/stable/index.html) and the 
        xarray (https://docs.xarray.dev/en/stable/index.html) packages

        Parameters
        ----------

        estimate : Literal["map", "mean"]
            Point estimate to return. 
            - map: Maximum a Posteriori. The sample that has the highest posterior probability.
              This sample considers the correlation structure of the posterior
            - mean: The average of all marginal parameter distributions.

        to : Literal["xarray", "dict"]
            Specifies the representation to transform the summarized data to. dict can
            be used to insert parameters in the .evaluate() method. While xarray is the
            standard view. Defaults to xarray

        Example
        -------

        >>> sim.best_estimate(to='dict')
        """
        if estimate == "mean":
            best_estimate = self.inferer.idata.posterior.mean(("chain", "draw"))

        elif estimate == "map":
            loglik = self.inferer.idata.log_likelihood\
                .sum(["id", "time"])\
                .to_array().sum("variable")
            
            sample_max_loglik = loglik.argmax(dim=("chain", "draw"))
            best_estimate = self.inferer.idata.posterior.sel(sample_max_loglik)
        else:
            raise GutsBaseError(
                f"Estimate '{estimate}' not implemented. Choose one of ['mean', 'map']"
            )


        if to == "xarray":
            return best_estimate
            
        elif to == "dict":
            return {k: v.values for k, v in best_estimate.items()}

        else:
            raise GutsBaseError(
                "PymobConverter.best_esimtate() supports only return types to=['xarray', 'dict']" +
                f"You used {to=}"
            )


    def evaluate(
        self, 
        parameters: Mapping[str, float|NumericArray|Sequence[float]] = {}, 
        y0: Mapping[str, float|NumericArray|Sequence[float]] = {}, 
        x_in: Mapping[str, float|NumericArray|Sequence[float]] = {}, 
    ):
        """Evaluates the model along the coordinates of the observations with given
        parameters, x_in, and y0. The dictionaries passed to the function arguments
        only overwrite the existing default parameters; which makes the usage very simple.

        Note that the first run of .evaluate() after calling the .dispatch_constructor()
        takes a little longer, because the model and solver are jit-compiled to JAX for
        highly efficient computations.

        Parameters
        ----------

        theta : Dict[float|Sequence[float]]
            Dictionary of model parameters that should be changed for dispatch.
            Unspecified model parameters will assume the default values, 
            specified under config.model_parameters.NAME.value

        y0 : Dict[float|Sequence[float]]
            Dictionary of initial values that should be changed for dispatch.
        
        x_in : Dict[float|Sequence[float]]
            Dictionary of model input values that should be changed for dispatch.


        Example
        -------

        >>> sim.dispatch_constructor()  # necessary if the sim object has been modified
        >>> # evaluate setting the background mortaltiy to zero
        >>> sim.evaluate(parameters={'hb': 0.0})  

        """
        evaluator = self.dispatch(theta=parameters, x_in=x_in, y0=y0)
        evaluator()
        return evaluator.results

    def load_exposure_scenario(
        self, 
        data: str|Dict[str,pd.DataFrame],
        sheet_name_prefix: str = "",
        rect_interpolate=False

    ):

        if isinstance(data, str):
            _data, time_unit = read_excel_file(
                path=data, 
                sheet_name_prefix=sheet_name_prefix,
                convert_time_to=self.unit_time
            )
        else:
            _data = data


        self._obs_backup = self.observations.copy(deep=True)

        # read exposure array from file
        exposure_dim = [
            d for d in self.config.data_structure.exposure.dimensions
            if d not in (self.config.simulation.x_dimension, self.config.simulation.batch_dimension)
        ]
        exposure = self._exposure_data_to_xarray(
            exposure_data=_data, 
            dim=exposure_dim[0]
        )

        # combine with observations
        new_obs = xr.combine_by_coords([
            exposure,
            self.observations.survival
        ]).sel(id=exposure.id)
        
        self.observations = new_obs.sel(time=[t for t in new_obs.time if t <= exposure.time.max()])
        self.config.simulation.x_in = ["exposure=exposure"]
        self.model_parameters["x_in"] = self.parse_input("x_in", exposure).ffill("time")  # type: ignore
        self.model_parameters["y0"] = self.parse_input("y0", drop_dims=["time"])

        self.dispatch_constructor()

    def export(self, directory: Optional[str] = None):
        self.config.simulation.skip_data_processing = False
        super().export(directory=directory)

class GutsSimulationConstantExposure(GutsBase):
    t_max = 10
    def initialize_from_script(self):
        self.config.data_structure.B = DataVariable(dimensions=["time"], observed=False)
        self.config.data_structure.D = DataVariable(dimensions=["time"], observed=False)
        self.config.data_structure.H = DataVariable(dimensions=["time"], observed=False)
        self.config.data_structure.survival = DataVariable(dimensions=["time"], observed=False)

        # y0
        self.config.simulation.y0 = ["D=Array([0])", "H=Array([0])", "survival=Array([1])"]
        self.model_parameters["y0"] = self.parse_input(input="y0", drop_dims=["time"])

        # parameters
        self.config.model_parameters.C_0 = Param(value=10.0, free=False)
        self.config.model_parameters.k_d = Param(value=0.9, free=True)
        self.config.model_parameters.h_b = Param(value=0.00005, free=True)
        self.config.model_parameters.b = Param(value=5.0, free=True)
        self.config.model_parameters.z = Param(value=0.2, free=True)

        self.model_parameters["parameters"] = self.config.model_parameters.value_dict
        self.config.simulation.model = "guts_constant_exposure"

        self.coordinates["time"] = np.linspace(0,self.t_max)

    def use_jax_solver(self):
        # =======================
        # Define model and solver
        # =======================

        self.coordinates["time"] = np.array([0,self.t_max])
        self.config.simulation.model = "guts_constant_exposure"

        self.solver = JaxSolver

        self.dispatch_constructor(diffrax_solver=Dopri5)

    def use_symbolic_solver(self):
        # =======================
        # Define model and solver
        # =======================

        self.coordinates["time"] = np.array([0,self.t_max])
        self.config.simulation.model = "guts_sympy"

        self.solver = mod.PiecewiseSymbolicSolver

        self.dispatch_constructor(diffrax_solver=Dopri5)


class GutsSimulationVariableExposure(GutsSimulationConstantExposure):
    t_max = 10
    def initialize_from_script(self):
        super().initialize_from_script()
        del self.coordinates["time"]
        exposure = create_artificial_data(
            t_max=self.t_max, dt=1, 
            exposure_paths=["topical"]
        ).squeeze()
        self.observations = exposure

        self.config.data_structure.exposure = DataVariable(dimensions=["time"], observed=True)

        self.config.simulation.x_in = ["exposure=exposure"]
        x_in = self.parse_input(input="x_in", reference_data=exposure, drop_dims=[])
        x_in = rect_interpolation(x_in=x_in, x_dim="time")
        self.model_parameters["x_in"] = x_in

        # parameters
        self.config.model_parameters.remove("C_0")

        self.model_parameters["parameters"] = self.config.model_parameters.value_dict
        self.config.simulation.solver_post_processing = "red_sd_post_processing"
        self.config.simulation.model = "guts_variable_exposure"


    def use_jax_solver(self):
        # =======================
        # Define model and solver
        # =======================

        self.model = self._mod.guts_variable_exposure
        self.solver = JaxSolver

        self.dispatch_constructor(diffrax_solver=Dopri5)

    def use_symbolic_solver(self, do_compile=True):
        # =======================
        # Define model and solver
        # =======================

        self.model = self._mod.guts_sympy
        self.solver = self._mod.PiecewiseSymbolicSolver

        self.dispatch_constructor(do_compile=do_compile, output_path=self.output_path)




from typing import List, Optional, Callable, Dict, Any, Tuple
from importlib import import_module
import warnings

import numpy as np
import pandas as pd
from expyDB.database_operations import create_database, experiment_to_db
from pymob.sim.config import dict_to_string

import glob
import os

import click
import pandas as pd

from expyDB.intervention_model import to_expydb, Experiment, PandasConverter
from guts_base.data.utils import datalad_locked_file_warning

def test_equality_of_exposure_patterns_in_treatment(df):
    for _, group in df.groupby("treatment_id"):
        exposures = group.pivot_table(
            # values="value", 
            index=["time", "treatment_id"], 
            columns="replicate_id"
        )

        equal_expo = exposures.values == exposures.values[:, ].reshape((-1, 1))
        if not np.all(equal_expo):
            raise RuntimeError(
                "Replicates in the same treatment ID have different exposure patterns."
            )
        
def create_new_columns_and_test_integrity_of_replicates(
    exposure, survival, n_reps, path
):
    assert np.all(exposure.columns == survival.columns)
    
    columns_new, treatment_reps = identify_replicates(frame=exposure)

    if not np.all(np.array(list(treatment_reps.values())) == n_reps):
        warnings.warn(
            f"Actual treatment replicates are different from "
            f"replicates ({n_reps}), given in Info sheet in file: "
            f"{path}"
        )

    return columns_new

def identify_replicates(frame):
    df = frame.drop(columns="time")

    # Find identical columns and assign group labels
    group_labels = {}
    used_cols = set()
    treatment_map = {}

    for col in df.columns:
        if col not in used_cols:
            # compare the column to each column of the dataframe with 
            # df.apply(func, axis=0) and get the column names
            identical_cols = df.columns[df.apply(lambda x: x.equals(df[col]), axis=0)].tolist()
            group_label = f'{len(group_labels) + 1}'
            group_labels[group_label] = identical_cols
            used_cols.update(identical_cols)

            for icol in identical_cols:
                treatment_map.update({icol: group_label})

    columns_new = [f"{treatment_map[col]}__{col}" for col in df.columns]
    treatment_reps = {key: len(cols) for key, cols in group_labels.items()}

    return columns_new, treatment_reps



def read_timeseries_sheet(path, sheet, sep=None):
    ts = pd.read_excel(path, sheet_name=sheet, index_col=0)  # type: ignore
    multi_index = pd.MultiIndex.from_tuples(
        [tuple(c.split(sep)) for c in ts.columns], names=["treatment_id", "timeseries_id"]
    )
    ts.columns = multi_index
    return ts


class OpenGutsIO:
    # TODO: Use preprocessing here and use map as a class attribute
    def __init__(self, file):
        self._file = file
        self.data = self.from_file(file)

    def _openguts_wide_to_long(self, frame, columns_new):
        frame_wide = frame.copy()

        frame_wide.columns = ["time"] + columns_new
        frame_long = pd.melt(
            frame=frame_wide, 
            id_vars=["time"], 
            value_vars=columns_new,
            var_name="exposure_id"
        )
        # create new index columns from new column names
        frame_long[["treatment_id", "replicate_id"]] = frame_long\
            .exposure_id.str.split("__", n=1, expand=True)
        frame_long = frame_long.drop(columns="exposure_id")
        
        return frame_long


    def _merge_tables(self, tables: List):
        data = tables.pop(0).set_index(["time", "treatment_id", "replicate_id"])

        for expo in tables:
            rdata =expo.set_index(["time", "treatment_id", "replicate_id"])
            data = pd.merge(
                left=data, 
                right=rdata, 
                how="left", 
                left_index=True, 
                right_index=True
            )

        return data

    def _read_timeseries(self, path, sheets):
        # design new columns based on the information about replicates and treatments
        timeseries_long_list = []
        timeseries_column_list = []
        time_units = {}
        for iv in sheets:
            timeseries_df = pd.read_excel(path, sheet_name=f"{iv}") 

            time_column = timeseries_df.columns[0]
            time_unit = time_column.lower().replace("time", "").strip(" []")

            # define replicates based on equality of columns
            timeseries_columns = [c for c in timeseries_df.columns[1:]]
            timeseries_long = self._openguts_wide_to_long(
                frame=timeseries_df, columns_new=timeseries_columns
            )
            intervention_long = timeseries_long.rename(columns={"value": iv})
            timeseries_long_list.append(intervention_long)
            timeseries_column_list.append(timeseries_columns)
            time_units.update({iv: time_unit})

        return self._merge_tables(timeseries_long_list).reset_index(), time_units


    def _read_openguts(self, path, metadata_sheetname="meta"):
        meta = pd.read_excel(path, sheet_name=metadata_sheetname, index_col=0).dropna(how="all")
        interventions = meta.loc["experiment__interventions","Value"]
        if interventions is None:
            raise ValueError("'experiment__interventions' must be defined in metadata")
        else:
            intervention_sheets = [i.strip("[]' ") for i in interventions.split(",")]  # type: ignore
        
        observations = meta.loc["experiment__observations","Value"]
        if observations is None:
            raise ValueError("'experiment__observations' must be defined in metadata")
        else:
            observation_sheets = [i.strip("[]' ") for i in observations.split(",")]  # type: ignore
        
        # survival_df = pd.read_excel(path, sheet_name="survival") 
        # survival_df = survival_df.rename(columns={"time [d]": "time"})

        # design new columns based on the information about replicates and treatments
        interventions_long, interventions_time_units = self._read_timeseries(path, intervention_sheets)
        observations_long, observations_time_units = self._read_timeseries(path, observation_sheets)
        time_unit = {
            "interventions": interventions_time_units,
            "observations": observations_time_units
        }

        # TODO test if all exposures within a treatment (replicates) were nominally the same 
        # test_equality_of_exposure_patterns_in_treatment(df=exposures_long)

        return interventions_long, observations_long, meta, time_unit

    def from_file(self, file) -> None:
        (
            interventions_long, 
            observations_long, 
            meta, 
            time_unit
        ) = self._read_openguts(path=file)

        self.interventions = interventions_long
        self.observations = observations_long
        self.time_unit = time_unit
        self.meta = meta

    def to_file(self, file):
        raise NotImplementedError(
            "This method should implement writing an excel file that corresponds"
            "to the original input file."
        )
    
    def to_experiment(self) -> Experiment:
        return Experiment.from_dict(data=dict(
            interventions=self.interventions, 
            observations=self.observations,
            meta=self.meta,
            time_units=self.time_unit
        ))

    def from_experiment(self, experiment: Experiment) -> None:
        data = experiment.to_dict()
        self.interventions=data["interventions"],
        self.observations=data["observations"],
        self.meta=data["meta"],
        self.time_units=data["time_units"],

    def to_xarray(self):
        return self.to_experiment().to_xarray()


def import_data_to_database(path, database, preprocessing: Optional[Callable] = None, preprocessing_out: Optional[str] = None):
    """This script takes raw data, preprocesses it to contain all 
    necessary metadata for expyDB. Then it creates an experiment Model and 
    processes adds it to the database
    """
    # preprocess path
    if preprocessing is not None:
        if preprocessing_out is None:
            filename = os.path.dirname(path)
            directory = os.path.basename(filename)
            new_path = path.replace(directory, f"processed_{directory}")
        else:
            filename = os.path.basename(path)
            new_path = preprocessing_out.format(filename=filename)

        os.makedirs(os.path.dirname(new_path), exist_ok=True)
        
        processed_path = preprocessing(path, new_path)
    else:
        processed_path = path

    # Preprocess excel to interventions and observations in Long form and a 
    # metadata Series as well as a default time unit
    openguts = OpenGutsIO(processed_path)
    
    # From excel to an Experiment Model instance
    experiment = openguts.to_experiment()

    # from the Model to the Database
    if not os.access(database, os.W_OK):
        warnings.warn(
            f"Did not write to database. The file '{database}' does "
            "not have write access."
        )
        return
    
    experiment.to_database(database=database)


def create_database_and_import_data_main(datasets_path, database_path, preprocessing=None, preprocessing_out=None):
    print("\n")
    print(f"Creating a database and importing data")
    print(f"======================================")

    if preprocessing is not None:
        module, func = preprocessing.rsplit(".", 1)
        mod = import_module(module)
        preprocessing_func = getattr(mod, func)
    else:
        preprocessing_func = None

    paths = []
    for p in datasets_path:
        if os.path.isfile(p):
            paths.append(p)
        else:
            paths.extend(glob.glob(os.path.join(p, "*.xlsx")))

    create_database(database=database_path, force=True)
    for p in paths:
        print(f"\nPreprocessing and importing file: {p}")
        import_data_to_database(
            path=p, database=database_path, 
            preprocessing=preprocessing_func,
            preprocessing_out=preprocessing_out
        )

@click.command()
@click.option("--datasets_path", type=str, multiple=True, help="The path to the directory where the excel files are located. Alternatively, use multiple times with paths to files")
@click.option("--database_path", type=str, help="The path to the database (should end with .db)")
@click.option("--preprocessing", type=str, help="Function used to preprocess the data", default=None)
@click.option("--preprocessing-out", type=str, help="A pattern that uses {filename} as a placeholder e.g. 'data/processed_data/{filename}. If unset, preprends 'processes_' to the dirname", default=None)
def create_database_and_import_data(datasets_path, database_path, preprocessing, preprocessing_out):
    create_database_and_import_data_main(
        datasets_path=datasets_path,
        database_path=database_path,
        preprocessing=preprocessing,
        preprocessing_out=preprocessing_out
    )


import numpy as np
import pytest
from guts_base.sim import GutsSimulationVariableExposure, GutsSimulationConstantExposure

@pytest.mark.slow
def test_guts_constant_exposure_with_symbolic():
    pytest.skip()
    sim = GutsSimulationConstantExposure()
    sim.initialize_from_script()

    sim.use_jax_solver()
    sim.dispatch_constructor(rtol=1e-7, atol=1e-8)
    evaluator = sim.dispatch(theta={})
    evaluator()
    
    sol_numerical = evaluator.results


    sim.use_symbolic_solver()
    evaluator = sim.dispatch(theta={})
    evaluator()

    sol_symbolic = evaluator.results

    diff = (
        sol_numerical.sel(time=[0, 180])
        - sol_symbolic.sel(time=[0, 180])
    )

    max_delta = np.abs(diff).max().to_array()
    np.testing.assert_array_less(max_delta, [1e-2, 1e-2, 1e-2])

    
@pytest.mark.slow
def test_guts_variable_exposure_with_symbolic():
    pytest.skip()
    sim = GutsSimulationVariableExposure()
    sim.initialize_from_script()

    sim.use_symbolic_solver(do_compile=True)
    evaluator = sim.dispatch(theta={})
    evaluator()
    sol_symbolic = evaluator.results

    sim.use_jax_solver()
    sim.dispatch_constructor(rtol=1e-7, atol=1e-8)
    evaluator = sim.dispatch(theta={})
    evaluator()
    sol_numerical = evaluator.results

    # make sure errors are small between exact solution and numerical solution
    # the errors come from:
    # a) integrating not exactly to t_eq
    # b) using a numerical switch in the ODE solution.
    diff = (
        sol_numerical.sel(time=np.arange(0,sim.t_max))
        - sol_symbolic.sel(time=np.arange(0,sim.t_max))
    )[["D", "H", "S"]]
    max_delta = np.abs(diff).max().to_array()
    np.testing.assert_array_less(max_delta, [5e-2, 5e-2, 5e-2])

    axes = sim._plot.plot_multiexposure(sol_numerical, vars=["exposure", "D", "H", "S"], color="tab:blue", label_prefix="ODE")
    axes = sim._plot.plot_multiexposure(sol_symbolic, vars=["exposure", "D", "H", "S"], axes=axes, color="tab:red", linestyle="--", label_prefix="exact")
    fig = axes[0].figure
    fig.savefig(f"{sim.output_path}/solution_comparison.png")


@pytest.mark.slow
def test_guts_variable_exposure_compiled():
    pytest.skip()
    sim = GutsSimulationVariableExposure()
    sim.initialize_from_script()

    sim.use_symbolic_solver(do_compile=False)
    evaluator = sim.dispatch(theta={})
    evaluator()
    sol_symbolic = evaluator.results

    sim.use_jax_solver()
    sim.dispatch_constructor(rtol=1e-7, atol=1e-8)
    evaluator = sim.dispatch(theta={})
    evaluator()
    sol_numerical = evaluator.results

    # make sure errors are small between exact solution and numerical solution
    # the errors come from:
    # a) integrating not exactly to t_eq
    # b) using a numerical switch in the ODE solution.
    diff = (
        sol_numerical.sel(time=np.arange(0,sim.t_max))
        - sol_symbolic.sel(time=np.arange(0,sim.t_max))
    )[["D", "H", "S"]]
    max_delta = np.abs(diff).max().to_array()
    np.testing.assert_array_less(max_delta, [5e-2, 5e-2, 5e-2])
